/**
 * Copyright (c) 2020, 2021 Jelle Hellings.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <functional>
#include <iostream>
#include <iterator>
#include <numeric>
#include <random>
#include "transaction_ostream.hpp"
#include "simulation_system.hpp"
#include "workload.hpp"


using namespace workload;

/**
 * Helper to print clock durations.
 */
std::ostream& operator<<(std::ostream& out, const clock_type::duration& duration)
{
    return out << std::chrono::duration_cast<std::chrono::microseconds>(duration).count();
}

/**
 * Helper to print medians.
 */
void print_half(std::ostream& out, const clock_type::duration& duration)
{
    out << std::chrono::duration_cast<std::chrono::microseconds>(duration).count() / 2;
}

template<class NumericType>
void print_half(std::ostream& out, const NumericType& value)
{
    out << value / 2;
}

/**
 * Obtain and print the median from the provided sequence.
 */
template<class Iterator>
void print_median(std::ostream& out, Iterator begin, Iterator end)
{
    auto hs = std::distance(begin, end) / 2;
    print_half(out, *std::next(begin, hs) + *std::prev(end, hs + 1));
}

/**
 * Print a boxplot [min, Q1, median, Q3, max] in which Q1 is the median of
 * the dataset up to (but excluding) the median and Q2 is the median of the
 * dataset starting after the median. Will print NA if a value is not available
 * (less than three elements in container).
 */
template<class Container>
void print_box_plot(std::ostream& out, Container container)
{
    std::sort(container.begin(), container.end());
    auto s = container.size();

    switch (s) {
        case 0:
            out << "NA\tNA\tNA\tNA\tNA";
            break;
        case 1:
        case 2:
            out << container.front() << '\t';
            print_median(out, container.cbegin(), container.cend());
            out << '\t';
            print_median(out, container.cbegin(), container.cend());
            out << '\t';
            print_median(out, container.cbegin(), container.cend());
            out << '\t' << container.back();
            break;
        default: {
            auto mid = s / 2;
            auto r = (s % 2 == 0) ? 1 : 0;
            auto mid_begin = std::next(container.cbegin(), mid - r);
            auto mid_end = std::next(mid_begin, r + 1);
            out << container.front() << '\t';
            print_median(out, container.cbegin(), mid_begin);
            out << '\t';
            print_median(out, container.cbegin(), container.cend());
            out << '\t';
            print_median(out, mid_end, container.cend());
            out << '\t' << container.back();
        }
    }
}

/**
 * Print the simulation report of a single simulation of ByShard. The scripts
 * are of the specified type, and the system had the specified number of shards.
 */
template<class... Args>
void print_report(const char* script_type, txn::shard_id_t num_shards, const simulation::report& report, Args&&... args)
{
    ((std::cout << args << '\t'), ...);

    /* Aggregrate totals from the report. */
    auto total_runtime = report.end - report.begin;
    auto total_steps = std::accumulate(report.script_steps.cbegin(), report.script_steps.cend(), std::size_t(0));
    auto total_duration = std::accumulate(report.script_times.cbegin(), report.script_times.cend(), clock_type::duration());

    /* General-purpose performance measurements. */
    std::cout << script_type << '\t' << num_shards << '\t'
              << total_runtime << '\t' << total_duration << '\t' << total_steps << '\t'
              << report.total_executed_steps << '\t' << report.total_executed_successful << '\t' << report.total_condition_failures << '\t'
              << report.total_forwards << '\t' << report.total_forwards_out_of_order << '\t';
              
    print_box_plot(std::cout, report.script_steps);
    std::cout << '\t';
    print_box_plot(std::cout, report.script_times);
    std::cout << '\t';
    print_box_plot(std::cout, report.shard_steps);
    std::cout << '\t';
    print_box_plot(std::cout, report.shard_steps_out_of_order);
    std::cout << '\t';

    /* Details specific to locks. */
    std::cout << report.total_wakeups << '\t'
              << report.total_read_locks << '\t' << report.total_write_locks << '\t'
              << report.failed_read_locks << '\t' << report.failed_write_locks << '\t';

    /* Details specific to voting. */
    std::cout << report.total_vote_waits << '\t' << report.total_vote_waits_positive << '\t'
              << report.total_yes_votes << '\t' << report.total_no_votes << '\t';


    /* Critical errors, should all be zero. */
    std::cout << report.total_failed_scripts << '\t'
              << report.total_remain_lock_waiting << '\t'
              << report.total_remain_vote_waiting << std::endl;
}

/**
 * Print the column headers for a report.
 */
template<class... Args>
 void print_report_headers(Args&&... args)
{
    ((std::cout << args << '\t'), ...);
    std::cout << "script_type" << '\t' << "num_shards" << '\t'
              << "total_runtime" << '\t' << "total_duration" << '\t' << "total_steps" << '\t'
              << "total_executed_steps" << '\t' << "total_executed_successful" << '\t' << "total_condition_failures" << '\t'
              << "total_forwards" << '\t' << "total_forwards_out_of_order" << '\t'
              << "script_steps_min" << '\t' << "script_steps_1Q" << '\t' << "script_steps_median" << '\t' << "script_steps_3Q" << '\t' << "script_steps_max" << '\t'
              << "script_times_min" << '\t' << "script_times_1Q" << '\t' << "script_times_median" << '\t' << "script_times_3Q" << '\t' << "script_times_max" << '\t'
              << "shard_steps_min" << '\t' << "shard_steps_1Q" << '\t' << "shard_steps_median" << '\t' << "shard_steps_3Q" << '\t' << "shard_steps_max" << '\t'
              << "shard_steps_out_of_order_min" << '\t' << "shard_steps_out_of_order_1Q" << '\t' << "shard_steps_out_of_order_median" << '\t' << "shard_steps_out_of_order_3Q" << '\t' << "shard_steps_out_of_order_max" << '\t'
              << "total_wakeups" << '\t'
              << "total_read_locks" << '\t' << "total_write_locks" << '\t'
              << "failed_read_locks" << '\t' << "failed_write_locks" << '\t'
              << "total_vote_waits" << '\t' << "total_vote_waits_positive" << '\t'
              << "total_yes_votes" << '\t' << "total_no_votes" << '\t'
              << "total_failed_scripts" << '\t' << "total_remain_lock_waiting" << '\t' << "total_remain_vote_waiting"
              << std::endl;
}


/**
 * Run the provided transaction workload on a system with the specified number
 * of accounts that are initialized using the provided generator, specified
 * number of shards, simulation costs for consensus and cluster-sending. For
 * each type of protocol enabled in protocols, the final report of the run will
 * be printed to standard output.
 */
template<class BalanceGenerator, class Protocols, class... Args>
void run_experiment(txn::account_id_t num_accounts, txn::shard_id_t num_shards,
                    const transactions& transactions, BalanceGenerator&& initial_balances,
                    const simulation::simulation_details& costs, const Protocols& protocols, Args&&... args)
{
    simulation::translate_system tsystem(num_accounts, num_shards);
    auto workloads = generate_scripts(transactions, tsystem, protocols);

    const char* names[] = {"lous", "los", "lru", "lruwf", "lrc", "lrcwf", "ls", "lswf",
                           "cous", "cos", "cruwf", "crcwf", "cswf",
                           "dous", "dos", "druwf", "drcwf", "dswf",
                           "ahl"}; 

    const char** label = names;
    for (auto it = workloads.cbegin(), end = workloads.cend(); it != end; ++label) {
        auto& workload = *it;
        ++it;

        if (!workload.empty()) {
            /* Only AHL---the last workload---requires a reference committee. */
            bool is_ahl = it == end;

            auto report = simulation::run(num_accounts, num_shards, costs, is_ahl,
                                          workload, std::forward<BalanceGenerator>(initial_balances));
            print_report(*label, num_shards, report, std::forward<Args>(args)...);
            std::cerr << '.';
        }
    }
}


/**
 * Scalability of the multi-shard transaction protocols: performance as
 * function of the number of shards.
 */
template<class InitialBalanceGenerator, class TransactionBalanceGenerator, class Protocols>
void experiment_scalability(const txn::account_id_t num_accounts, const txn::script_id_t num_scripts,
                            InitialBalanceGenerator&& initial_balances, TransactionBalanceGenerator&& transaction_balances,
                            const simulation::simulation_details costs, const Protocols& protocols)
{
    std::cout << "experiment_scalability" << std::endl;
    print_report_headers();
    std::random_device rd;
    std::uniform_int_distribution transaction_accounts(0, (int)num_accounts - 1);
    auto transactions = generate_transactions(num_scripts, check_transfer(), true,
                                              make_constant_generator(8), make_constant_generator(4), make_constant_generator(4),
                                              std::bind(transaction_accounts, std::ref(rd)), transaction_balances);

    for (auto num_shards : {1, 2, 4, 8, 16, 32, 64, 128, 256, 512}) {
        run_experiment(num_accounts, num_shards, transactions, initial_balances, costs, protocols);
    }
    std::cout << std::endl;
}

/**
 * Impact of contention: performance as a function of the number of accounts.
 * The more accounts, the less likely two transactions affect the same account.
 */
template<class InitialBalanceGenerator, class TransactionBalanceGenerator, class Protocols>
void experiment_contention(const txn::script_id_t num_scripts, const txn::shard_id_t num_shards,
                           InitialBalanceGenerator&& initial_balances, TransactionBalanceGenerator&& transaction_balances,
                           const simulation::simulation_details costs, const Protocols& protocols)
{
    std::cout << "experiment_contention" << std::endl;
    print_report_headers("account_factor");
    std::random_device rd;
    for (auto account_factor : {1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096}) {
        const txn::account_id_t num_accounts = num_shards * account_factor;
        std::uniform_int_distribution transaction_accounts(0, (int)num_accounts - 1);
        auto transactions = generate_transactions(num_scripts, check_transfer(), true,
                                                  make_constant_generator(8), make_constant_generator(4), make_constant_generator(4),
                                                  std::bind(transaction_accounts, std::ref(rd)), transaction_balances);
        run_experiment(num_accounts, num_shards, transactions, initial_balances, costs, protocols, account_factor);
    }
    std::cout << std::endl;
}

/**
 * Scalability of the multi-shard transaction protocols: performance as
 * function of the number of shards.
 */
template<class InitialBalanceGenerator, class TransactionBalanceGenerator, class Protocols>
void experiment_factor_scalability(const txn::account_id_t num_accounts, const txn::script_id_t num_scripts,
                                   InitialBalanceGenerator&& initial_balances, TransactionBalanceGenerator&& transaction_balances,
                                   const simulation::simulation_details costs, const Protocols& protocols)
{
    std::cout << "experiment_factor_scalability" << std::endl;
    print_report_headers("scale_factor");
    std::random_device rd;

    for (auto scale_factor : {1, 2, 4, 8, 16, 32, 64, 128, 256, 512}) {
        std::uniform_int_distribution transaction_accounts(0, (int)(num_accounts * scale_factor - 1));
        auto transactions = generate_transactions(num_scripts, check_transfer(), true,
                                                  make_constant_generator(8), make_constant_generator(4), make_constant_generator(4),
                                                  std::bind(transaction_accounts, std::ref(rd)), transaction_balances);
        run_experiment(num_accounts * scale_factor, scale_factor, transactions, initial_balances, costs, protocols, scale_factor);
    }
    std::cout << std::endl;
}


/**
 * Main entry point. Run all the experiments.
 */
int main(int argc, char* argv[])
{
    using namespace std::chrono_literals;
    std::random_device rd;

    const txn::account_id_t num_accounts = 512 * 16;
    const txn::script_id_t num_scripts = 5000;
    const txn::shard_id_t num_shards = 64;
    auto simulation_costs = simulation::simulation_details::pbft(10ms, 1000);
    auto protocols = protocols_all();

    auto initial_balances = make_constant_generator(2000);
    std::binomial_distribution<txn::balance_t> transaction_balance_distribution(1000, 0.5);
    auto transaction_balances = std::bind(transaction_balance_distribution, std::ref(rd));

    experiment_scalability(num_accounts, num_scripts, initial_balances, transaction_balances, simulation_costs, protocols);
    experiment_contention(num_scripts, num_shards, initial_balances, transaction_balances, simulation_costs, protocols);
    experiment_factor_scalability(num_accounts, num_scripts, initial_balances, transaction_balances, simulation_costs, protocols);
}