/**
 * Copyright (c) 2020 Jelle Hellings.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_MESSAGE_QUEUE_HPP
#define INCLUDE_MESSAGE_QUEUE_HPP
#include <condition_variable>
#include <deque>
#include <mutex>
#include <optional>
#include "utility_container.hpp"


/**
 * This class implements a multi-threaded message queue that is designed for
 * situations in which multiple producers send values of type T and
 * multiple consumers receive these values.
 */
template<class T, class Container = std::deque<T>>
class message_queue
{
public:
    using value_type = T;
    using container_type = Container;
    using size_type = typename container_type::size_type;


    /**
     * Return the nmber of messages in the queue.
     */
    size_type size() const
    {
        std::lock_guard<std::mutex> lock(data_mutex);
        return data.size();
    }

    /**
     * Return trur if the queue is empty.
     */
    bool empty() const
    {
        return size() == 0;
    };

    /**
     * Construct and send a new message.
     */
    template<class... Args> void send(Args&&... args)
    {
        std::lock_guard<std::mutex> lock(data_mutex);
        data.emplace_back(std::forward<Args>(args)...);
        empty_cv.notify_one();
    }

    /**
     * Receive a message, will wait until one is available.
     */
    value_type receive()
    {
        std::unique_lock<std::mutex> lock(data_mutex);

        /* Do not call this->size() here, would yield deadlock. */
        while (data.size() == 0) {
            empty_cv.wait(lock);
        }
        return util::front_pop(data);
    }

    /**
     * Receive a message, if one is available.
     */
    std::optional<value_type> try_receive()
    {
        std::lock_guard<std::mutex> lock(data_mutex);

        /* Do not call this->size() here, would yield deadlock. */
        if (data.size() == 0) {
            return {};
        }
        return util::front_pop(data);
    }


    message_queue() = default;
    message_queue(message_queue&&) = default;
    message_queue& operator=(message_queue&&) = default;
    message_queue(const message_queue&) = delete;
    message_queue& operator=(const message_queue&) = delete;


private:
    /* The raw queue used to send and receive messages. */
    std::deque<value_type> data;

    /* Mutex used to provide thread-safe access to data. The mutex needs to be
     * mutable to be able to lock in const member functions. */
    mutable std::mutex data_mutex;

    /* Variable on which threads can wait that try to receive from a
     * currently-empty queue.*/
    std::condition_variable empty_cv;
};

#endif