/**
 * Copyright (c) 2020, 2021 Jelle Hellings.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "simulation_system.hpp"

#include <algorithm>
#include <functional>
#include <thread>
#include "transaction_replica.hpp"
#include "transaction_utility.hpp"
#include "utility_container.hpp"


using namespace simulation;


shard_id_account_id_pair simulation::account_assignment(txn::account_id_t num_accounts, txn::shard_id_t num_shards, txn::account_id_t global_account_id)
{
    auto q = num_accounts / num_shards;
    auto r = num_accounts % num_shards;

    /**
     * The first r shards each get `q + 1' accounts, the other shards get `q'
     * accounts. Hence, we have `q * num_shards + r == num_accounts' accounts
     * assigned to shards.
     */
    auto first_rs = r * (q + 1);
    if (global_account_id < first_rs) {
        auto shard_id = global_account_id / (q + 1);
        auto account_id = global_account_id % (q + 1);
        return {shard_id, account_id};
    }
    else {
        auto shard_id = r + ((global_account_id - first_rs) / q);
        auto account_id = (global_account_id - first_rs) % q;
        return {shard_id, account_id};
    }
}

account_id_range simulation::account_range_assignment(txn::account_id_t num_accounts, txn::shard_id_t num_shards, txn::shard_id_t shard_id)
{
    auto q = num_accounts / num_shards;
    auto r = num_accounts % num_shards;

    /**
     * The first r shards each get `q + 1' accounts, the other shards get `q'
     * accounts. Hence, we have `q * num_shards + r == num_accounts' accounts
     * assigned to shards.
     */
    if (shard_id < r) {
        return {(q + 1) * shard_id, (q + 1) * (shard_id + 1)};
    }
    else
    {
        return {q * shard_id + r, q * (shard_id + 1) + r};
    }
}


/* Implementation of translate_system. */

txn::script_id_t translate_system::new_script_id(const txn::operation_list&)
{
    return sequence_script_id++;
}

std::optional<txn::shard_id_t> translate_system::account_to_shard(txn::account_id_t account_id) const
{
    if (account_id < num_accounts) {
        return account_assignment(num_accounts, num_shards, account_id).shard_id;
    }
    else {
        return {};
    }
}

txn::shard_id_t simulation::translate_system::reference_committee_ahl_id() const
{
    return num_shards;
}


/* Implementation for deriving reports (details::run_helper). */

/**
 * Order a (script, time) pair.
 */
bool script_time_order(const txn::script_time& lhs, const txn::script_time& rhs)
{
    return (lhs.script_id < rhs.script_id) ||
           (lhs.script_id == rhs.script_id && lhs.time_point < rhs.time_point);
}

/**
 * Collect the general-purpose measurements and update the raport.
 */
void collect_measurement(report& report, const txn::details::measurement& data,
                         std::vector<txn::script_time>& begins, std::vector<txn::script_time>& ends)
{
    report.total_failed_scripts += data.failed_scripts.size();
    report.total_executed_steps += data.script_begins.size();
    report.total_executed_successful += data.executed_successful;

    /* Collect all begin and end time points. */
    begins.insert(begins.end(), data.script_begins.cbegin(), data.script_begins.cend());
    ends.insert(ends.end(), data.script_ends.cbegin(), data.script_ends.cend());
}

/**
 * Construct a report out of the provided state machine data.
 */
report construct_report(const std::vector<txn::account_state_machine<replica_system>>& account_state_machines,
                        const std::vector<txn::vote_state_machine<replica_system>>& vote_state_machines)
{
    report result;

    /* Collect the basic performance numbers and build helper structures to
     * compute script_steps and script_times. */
    std::vector<txn::script_time> begins;
    std::vector<txn::script_time> ends;

    /* Data from the account state machines. */
    for (auto& data : account_state_machines) {
        collect_measurement(result, data, begins, ends);
        result.total_condition_failures += data.condition_failures;
        result.total_forwards += data.total_forwards;
        result.total_forwards_out_of_order += data.total_forwards_out_of_order;

        result.total_wakeups += data.script_wakeups.size();
        result.total_read_locks += data.total_read_locks;
        result.total_write_locks += data.total_write_locks;
        result.failed_read_locks += data.failed_read_locks;
        result.failed_write_locks += data.failed_write_locks;
        result.shard_steps.emplace_back(data.script_begins.size());

        /* Iterate over every account to check for stuck transactions, these
         * are bugs if they exist. */
        for (auto& account : data.accounts) {
            result.total_remain_lock_waiting += account.lock_wait_queue.size();
        }
    }

    /* Data from the vote state machines. */
    for (auto& data : vote_state_machines) {
        collect_measurement(result, data, begins, ends);

        result.total_vote_waits += data.total_vote_waits;
        result.total_vote_waits_positive += data.total_vote_waits_positive;
        result.total_yes_votes += data.total_yes_votes;
        result.total_no_votes += data.total_no_votes;
        result.shard_steps_out_of_order.emplace_back(data.script_begins.size());

        /* Check for stuck transactions, these are bugs if they exist. */
        result.total_remain_vote_waiting = data.vote_wait_queue.size();
    }

    /* Process the collected script time points. */
    std::sort(begins.begin(), begins.end(), script_time_order);
    std::sort(ends.begin(), ends.end(), script_time_order);

    auto it = begins.cbegin(), end = begins.cend();
    auto ends_begin = ends.cbegin(), ends_it = ends.cbegin(), ends_end = ends.cend();
    while (it != end) {
        /* This is the first entry for this script in begins and might be the
         * first time point for the first script. */
        auto begin_point = it->time_point;
        result.begin = std::min(result.begin, begin_point);

        /* Compute the number of steps this script has taken (based on the
         * number of entries in begins). */
        std::size_t num_steps = 0;
        auto script_id = it->script_id;
        while (it != end && it->script_id == script_id) {
            ++it;
            ++num_steps;
        }

        /* Search for the last entry in ends with at most the specified
         * script_id. */
        while (ends_it != ends_end && ends_it->script_id <= script_id) {
            ++ends_it;
        }

        /* Check whether we found an entry in ends with exactly the specified
         * script_id. */
        if (ends_it != ends_begin && std::prev(ends_it)->script_id == script_id) {
            /* This is the last entry for this script in ends and might be the
             * last time point for the last script. */
            auto end_point = std::prev(ends_it)->time_point;
            result.end = std::max(result.end, end_point);

            /* Register the runtime and the number of steps. */
            result.script_steps.emplace_back(num_steps);
            result.script_times.emplace_back(end_point - begin_point);
        }
    }

    return result;
}

/* Implementation of details::run_helper. */

report details::run_helper(txn::account_id_t num_accounts, txn::shard_id_t num_shards,
                           const simulation_details& costs, bool reference_committee_ahl,
                           const workload::scripts& workload, const std::vector<txn::balance_t>& balances)
{
    /* Set up the simulated replica system. */
    replica_system rsystem(num_accounts, num_shards, costs, reference_committee_ahl, workload);

    /* Construct the script that termination execution of state machines. */
    txn::operation_list termination = {txn::operation_forceshutdown()};
    txn::script termination_script{txn::script::system_script_id, 0, termination};
    txn::script_execution termination_execution{&termination_script, 0};
    replica_system::timed_message termination_message{termination_execution, clock_type::now()};

    /* Set up each individual replica (simulated shard). */
    std::vector<txn::account_state_machine<replica_system>> account_state_machines;
    std::vector<txn::vote_state_machine<replica_system>> vote_state_machines;
    while (account_state_machines.size() != num_shards) {
        auto shard_id = account_state_machines.size();
        auto pair = account_range_assignment(num_accounts, num_shards, shard_id);
        account_state_machines.emplace_back(rsystem, shard_id, balances.begin() + pair.begin, balances.begin() + pair.end);
        vote_state_machines.emplace_back(rsystem, shard_id);
    }
    if (reference_committee_ahl) {
        auto shard_id = account_state_machines.size();
        account_state_machines.emplace_back(rsystem, shard_id, balances.end(), balances.end());
        vote_state_machines.emplace_back(rsystem, shard_id);
    }

    /* Start up each individual replica (simulate a shard) .*/
    std::vector<std::thread> shard_threads;
    for (auto& state_machine : account_state_machines) {
        shard_threads.emplace_back(txn::main_loop<txn::account_state_machine<replica_system>>, std::ref(state_machine));
    }
    for (auto& state_machine : vote_state_machines) {
        shard_threads.emplace_back(txn::main_loop<txn::vote_state_machine<replica_system>>, std::ref(state_machine));
    }

    /* Wait until all operations have finished. */
    while (!rsystem.simulation_finished()) {
        std::this_thread::sleep_for(std::chrono::seconds(1));
    }

    /* Send the shutdown signal to all channels. */
    for (auto& channel : rsystem.channels) {
        channel.send(termination_message);
    }
    for (auto& channel : rsystem.channels_out_of_order) {
        channel.send(termination_message);
    }

    /* Wait until all shards have terminated. */
    for (auto& thread : shard_threads) {
        thread.join();
    }

    /* Return the report. */
    return construct_report(account_state_machines, vote_state_machines);
}


/* Implementation of simulation_details. */

simulation_details simulation_details::pbft(std::chrono::microseconds message_delay, std::size_t rate)
{
    using namespace std::chrono_literals;
    return simulation_details{3 * message_delay,
                              std::chrono::duration_cast<std::chrono::microseconds>(1s) / rate,
                              message_delay};
}

simulation_details simulation_details::hotstuff(std::chrono::microseconds message_delay)
{
    return simulation_details{8 * message_delay,
                              2 * message_delay,
                              message_delay };
}



/* Implementation of replica_system. */

txn::script_execution replica_system::receive_execution(replica_id_type replica_id)
{
    /* Earliest point at which the next consensus step is finished. */
    auto& tick = replica_tick[replica_id];
    tick += costs.consensus_tick;

    /* First, obtain a next message available to reach consensus on. */
    if (!replica_next_received[replica_id].has_value()) {
        replica_next_received[replica_id] = channels[replica_id].try_receive();
    }

    /* If there is a next message and there is a new transaction in the
     * workload, then return the message if consensus is finished before the
     * next tick. Otherwise, return a new transaction from the workload. */
    if (!client_workload[replica_id].empty()) {
        auto now = clock_type::now();
        if (tick > now) {
            std::this_thread::sleep_for(tick - now);
        }
        if (!replica_next_received[replica_id].has_value() ||
                replica_next_received[replica_id].value().arrival_time > tick) {
            return {util::front_pop(client_workload[replica_id]), 0};
        }
        else {
            auto execution = replica_next_received[replica_id].value().message;
            replica_next_received[replica_id].reset();
            return execution;
        }
    }
    else {
        if (replica_next_received[replica_id].has_value()) {
            auto value = replica_next_received[replica_id].value();
            replica_next_received[replica_id].reset();

            tick = std::max(tick, value.arrival_time);
            auto now = clock_type::now();
            if (tick > now) {
                std::this_thread::sleep_for(tick - now);
            }
            return value.message; 
        }
        else {
            auto value = channels[replica_id].receive();

            tick = std::max(tick, value.arrival_time);
            auto now = clock_type::now();
            if (tick > now) {
                std::this_thread::sleep_for(tick - now);
            }
            return value.message; 
        }
    }
}

txn::script_execution simulation::replica_system::receive_execution_out_of_order(replica_id_type replica_id)
{
    auto value = channels_out_of_order[replica_id].receive();
    auto now = clock_type::now();
    if (value.arrival_time > now) {
        std::this_thread::sleep_for(value.arrival_time - now);
    }
    return value.message;
}

bool replica_system::forward_execution(replica_id_type from_replica_id, txn::script_id_t to_shard_id, const txn::script_execution& execution)
{
    if (to_shard_id < true_num_shards) {
        ++active_steps;
        if (to_shard_id == from_replica_id) {
            channels[to_shard_id].send(timed_message{execution, clock_type::now() + costs.consensus_cost});
        }
        else {
            channels[to_shard_id].send(timed_message{execution, clock_type::now() + costs.consensus_cost + costs.cluster_sending_cost});
        }
        return true;
    }
    else {
        return false;
    }
}

bool replica_system::forward_execution_out_of_order(replica_id_type from_replica_id, txn::script_id_t to_shard_id, const txn::script_execution& execution)
{
    if (to_shard_id < true_num_shards) {
        ++active_steps;
        if (to_shard_id == from_replica_id) {
            channels_out_of_order[to_shard_id].send(timed_message{execution, clock_type::now()});
        }
        else {
            channels_out_of_order[to_shard_id].send(timed_message{execution, clock_type::now() + costs.cluster_sending_cost});
        }
        return true;
    }
    else {
        return false;
    }
}


std::optional<txn::account_id_t> replica_system::local_account_id(replica_id_type from_replica_id, txn::account_id_t global_account_id) const
{
    if (global_account_id < num_accounts) {
        auto pair = account_assignment(num_accounts, num_shards, global_account_id);
        if (from_replica_id == pair.shard_id) {
            return pair.account_id;
        }
    }
    return {};
}

void simulation::replica_system::release_execution(replica_id_type replica_id, const txn::script_execution& execution)
{
    --active_steps;
}

replica_system::replica_system(txn::account_id_t num_accounts, txn::shard_id_t num_shards,
                               const simulation_details& costs, bool reference_committee_ahl,
                               const workload::scripts& workload) :
                                    num_accounts(num_accounts),
                                    num_shards(num_shards),
                                    true_num_shards(num_shards + ((reference_committee_ahl) ? 1 : 0)),
                                    costs(costs),
                                    channels(true_num_shards),
                                    channels_out_of_order(true_num_shards),
                                    client_workload(true_num_shards),
                                    active_steps(0),
                                    replica_next_received(true_num_shards),
                                    replica_tick(true_num_shards, clock_type::now())
{
    /* Set up the client workload. */
    for (auto& script : workload) {
        if (script.initial_shard < true_num_shards) {
            client_workload[script.initial_shard].emplace_back(&script);
            ++active_steps;
        }
    }
}


bool replica_system::simulation_finished() const
{
    return active_steps.load() == 0;
}