/**
 * Copyright (c) 2020, 2021 Jelle Hellings.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_SIMULATION_SYSTEM_HPP
#define INCLUDE_SIMULATION_SYSTEM_HPP
#include <atomic>
#include <chrono>
#include <deque>
#include <optional>
#include <utility>
#include <vector>
#include "clock.hpp"
#include "message_queue.hpp"
#include "workload.hpp"


namespace simulation
{
    /**
     * A (shard identifier, account identifier)-pair.
     */
    struct shard_id_account_id_pair
    {
        txn::shard_id_t shard_id;
        txn::account_id_t account_id;
    };

    /**
     * A range [begin, end) of account identifiers.
     */
    struct account_id_range
    {
        txn::account_id_t begin;
        txn::account_id_t end;
    };


    /**
     * A simulated sharded system with num_accounts accounts distributed over
     * num_shards will assign each shard an equal share of accounts (up to
     * rounding). The following function will take a global account identifier
     * and return both the shard that account belongs to and the local
     * identifier of that account within the shard.
     *
     * We assume that global_account_id < num_accounts.
     */
    shard_id_account_id_pair account_assignment(txn::account_id_t num_accounts, txn::shard_id_t num_shards, txn::account_id_t global_account_id);

    /**
     * Return the range [begin, end) such that all accounts with identifier i,
     * begin <= i < end, are assigned to the shard with identifier shard_id.
     *
     * We assume that shard_id < num_shards.
     */
    account_id_range account_range_assignment(txn::account_id_t num_accounts, txn::shard_id_t num_shards, txn::shard_id_t shard_id);

    /**
     * Provide the necessary details on a simulated sharded system for the
     * purpose of translating transactions into scripts.
     */
    class translate_system
    {
    public:
        /* Configuration parameters of the system. */
        const txn::account_id_t num_accounts;
        const txn::shard_id_t num_shards;


        /**
         * Generate a script identifier that can be used by the system to
         * identify the script holding the provided operations. This system
         * simply returns a monotonically increasing counter.
         */
        txn::script_id_t new_script_id(const txn::operation_list&);

        /**
         * Return the identifier of the shard the account with the specified
         * account identifier is assigned to.
         */
        std::optional<txn::shard_id_t> account_to_shard(txn::account_id_t account_id) const;

        /**
         * Return the shard identifier of the reference committee (AHL).
         */
        txn::shard_id_t reference_committee_ahl_id() const;


        /**
         * Construct the translate system
         */
        translate_system(txn::account_id_t num_accounts, txn::shard_id_t num_shards) :
                num_accounts(num_accounts), num_shards(num_shards), sequence_script_id(0) { }


    private:
        /* Used for generating script identifiers. */
        txn::script_id_t sequence_script_id;
    };


    /**
     * Performance report returned after completing a simulation run.
     */
    struct report
    {
        /* Total number of executed script steps and failed scripts. */
        std::size_t total_failed_scripts = 0;
        std::size_t total_executed_steps = 0;

        /* Number of scripts executed succesful. */
        std::size_t total_executed_successful = 0;

        /* Condition failures. */
        std::size_t total_condition_failures = 0;

        /* Total amount of communication. */
        std::size_t total_forwards = 0;
        std::size_t total_forwards_out_of_order = 0;

        /* Performance statistics on read and write locks. */
        std::size_t total_wakeups = 0;
        std::size_t total_read_locks = 0;
        std::size_t total_write_locks = 0;
        std::size_t failed_read_locks = 0;
        std::size_t failed_write_locks = 0;
        std::size_t total_remain_lock_waiting = 0;

        /* Vote-waits, yes votes, and no votes. */
        std::size_t total_vote_waits = 0;
        std::size_t total_vote_waits_positive = 0;
        std::size_t total_yes_votes = 0;
        std::size_t total_no_votes = 0;
        std::size_t total_remain_vote_waiting = 0;

        /* Data on individual scripts: script_steps[i] holds the number of steps
         * the i-th transaction performed and script_time[i] holds the total
         * runtime of that same transaction. */
        std::vector<std::size_t> script_steps = {};
        std::vector<clock_type::duration> script_times = {};

        /* Total number of steps per shard. */
        std::vector<std::size_t> shard_steps = {};
        std::vector<std::size_t> shard_steps_out_of_order = {};

        /* Total runtime of the simulation determined by the start of the first
         * script and end of the last script */
        clock_type::time_point begin = clock_type::time_point::max();
        clock_type::time_point end = clock_type::time_point::min();
    };


    /**
     * The simulation simulates consensus steps and cluster-sending steps. To
     * capture the costs associated with these steps---which heavily depend on
     * details of the used protocols---we use a few parameters.
     */
    struct simulation_details
    {
        /* The minimal duration of a single consensus step measured from the
         * point of proposal of a decision to the point where replicas fully
         * accepted the decision. */
        std::chrono::microseconds consensus_cost;

        /* The minimal time between consecutive accepted consensus decisions.
         * Many protocols can either make several consensus decisions in
         * parallel (e.g., PBFT when given sufficient memory buffers at each 
         * replica) or overlap steps of consecutive consensus decisions (e.g.,
         * event-based HotStuff). */
        std::chrono::microseconds consensus_tick;

        /* The minimal duration of a single cluster-sending step. Here, we
         * assume reliable communication, in which case the cluster-sending
         * steps required for sharding can be done in one message delay. */
        std::chrono::microseconds cluster_sending_cost;

        /**
         * Return the consensus parameters for PBFT. The minimal duration of a
         * single consensus decision is fully determined by the message delay,
         * as each consensus decision goes through three phases of message
         * exchange.
         *
         * As PBFT supports out-of-order processing of consensus decisions if
         * sufficient memory buffers are available at each replica, the rate at
         * which decisions can be made is determined by network bandwidth. We
         * use this rate as an input parameter.
         */
        static simulation_details pbft(std::chrono::microseconds message_delay, std::size_t rate);

        /**
         * Return the consensus parameters for HotStuff (event-based). The
         * minimal duration for a single consensus decision is fully determined
         * by the message delay, as each consensus goes through four phases---
         * each consisting of two steps of message exchange (one primary-to-all,
         * one all-to-next-primary). In event-based HotStuff, consensus
         * decisions can be started every phase, but only after the next-primary
         * of the next phase learns of the decisions of the current phase.
         * Hence, the protocol is entirely bounded by message delays. 
         */
        static simulation_details hotstuff(std::chrono::microseconds message_delay);
    };


    namespace details
    {
        /* Internal helper for implementing simulation::run(). */
        report run_helper(txn::account_id_t num_accounts, txn::shard_id_t num_shards,
                          const simulation_details& simulation_details, bool reference_committee_ahl,
                          const workload::scripts& workload, const std::vector<txn::balance_t>& balances);
    };


    /**
     * Provide the necessary details on a simulated sharded system for the
     * purpose of running a replica state machine.
     */
    class replica_system
    {
    public:
        /* Configuration parameters of the system. */
        const txn::account_id_t num_accounts;
        const txn::shard_id_t num_shards;

        /* True number of shards; which includes the reference committee that
         * is only enabled for AHL. */
        const txn::shard_id_t true_num_shards;

        /* Configuration parameters of the simulation. */
        const simulation_details costs;

        /* In a simulated system, each shard has one replica. */
        using replica_id_type = txn::shard_id_t;


        /**
         * Receive a script execution for the specified replica.
         */
        txn::script_execution receive_execution(replica_id_type replica_id);
        txn::script_execution receive_execution_out_of_order(replica_id_type replica_id);

        /**
         * Send a script execution from the specified replica to the specified
         * shard. Returns true if that was a valid operation.
         */
        bool forward_execution(replica_id_type from_replica_id, txn::script_id_t to_shard_id, const txn::script_execution& execution);
        bool forward_execution_out_of_order(replica_id_type from_replica_id, txn::script_id_t to_shard_id, const txn::script_execution& execution);

        /**
         * Map a global account identifier to the local identifier used to
         * identify that account in the specified replica (if it is held by that
         * replica).
         */
        std::optional<txn::account_id_t> local_account_id(replica_id_type from_replica_id,
                                                          txn::account_id_t global_account_id) const;

        /**
         * Register the end of execution of the specified script execition by
         * the specified replica.
         */
        void release_execution(replica_id_type replica_id, const txn::script_execution& execution);


        replica_system(const replica_system&) = delete;
        replica_system& operator=(const replica_system&) = delete;
        replica_system(replica_system&&) = delete;
        replica_system& operator=(replica_system&&) = delete;


    private:
        /**
         * For each message, we precompute the arrival time after which the
         * shard can optimistically process the message. Here, we use the
         * assumption that consensus steps can be performed out-of-order:
         * consensus to process a message in round r can be reached before
         * preceding rounds are finished. This optimistic assumption holds for
         * protocols such as PBFT, but not for protocols such as HotStuff that
         * replace leaders every round.
         */
        struct timed_message
        {
            txn::script_execution message;
            clock_type::time_point arrival_time;
        };

        using const_script_pointer = const txn::script*;

        /* Communication channels between shards (replicas), channel[i] is used
         * to send messages to the i-th shard (replica). */
        std::vector<message_queue<timed_message>> channels;

        /* Communication channels between shards for out-of-order messages
         * (votes), channel_out_of_order[i] is used to send messages to the i-th
         * shard (replica). */
        std::vector<message_queue<timed_message>> channels_out_of_order;

        /* Communication channel between shards (replicas) and clients,
         * client_workload[i] points to all the workload scripts that clients
         * will send to the i-th shard (replica) as part of the simulated
         * workload. These scripts will start execution at offset 0. */
        std::vector<std::deque<const_script_pointer>> client_workload;

        /* The system keeps track of the active steps: the number of internal
         * steps that are still to-be-processed: initially, this number is the
         * number of valid client scripts in workload. During execution, the
         * number can increase if executions are forwarded and decrease if
         * executions end at a replica. Using this count, we can determine when 
         * to terminate the simulation. */
        std::atomic_uint64_t active_steps;

        /* To maximize performance, replicas will process a new client
         * transaction whenever consensus on the next-received message takes
         * more time than finishing the next consensus tick. For replica i, we
         * store such next-received message in replica_next_received[i], making
         * it available for future lock-free retrieval, and we store the last
         * consensus tick in replica_tick[i]. */
        std::vector<std::optional<timed_message>> replica_next_received;
        std::vector<clock_type::time_point> replica_tick;

        /**
         * Construct the replica system based on the provided translate_system.
         * The system simulates consensus steps of the specified cost and will
         * assign an initial client workload based on the provided workload.
         */
        replica_system(txn::account_id_t num_accounts, txn::shard_id_t num_shards,
                       const simulation_details& costs, bool reference_committee_ahl,
                       const std::vector<txn::script>& workload);

        /**
         * Return true if the simulation finished processing the entire
         * workload.
         */
        bool simulation_finished() const;


        friend report details::run_helper(txn::account_id_t, txn::shard_id_t, const simulation_details&, bool,
                                          const std::vector<txn::script>&, const std::vector<txn::balance_t>&);
    };


    /**
     * Run a simulation on the specified workload. The initial account balanace
     * of the account with identifier account_id will obtained via a call
     * balance_f(account_id).
     */
    template<class BalanceGen>
    report run(txn::account_id_t num_accounts, txn::shard_id_t num_shards,
               const simulation_details& costs, bool reference_committee_ahl,
               const workload::scripts& workload, BalanceGen&& balance_f)
    {
        /* Generate the initial balances. */
        std::vector<txn::balance_t> balances;
        while (balances.size() < num_accounts) {
            balances.emplace_back(balance_f(balances.size()));
        }
        return details::run_helper(num_accounts, num_shards, costs,
                                   reference_committee_ahl, workload, balances);
    }
};

#endif