/**
 * Copyright (c) 2020 Jelle Hellings.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "transaction_assembler.hpp"

#include <string>


using namespace txn;


/* Implementation of assembler. */

assembler::fragment assembler::create_fragment()
{
    fragments.emplace_back();
    return {fragments.back(), fragments.size() - 1};
}

offset_t assembler::front_jump_label(const fragment& fragment, offset_t offset)
{
    if (fragments.size() <= fragment.fragment_id) {
        throw invalid_fragment_error(fragment.fragment_id);
    }
    jump_labels.emplace_back(internal_jump_label{fragment.fragment_id, offset, true});
    return jump_labels.size() - 1;
}

offset_t assembler::front_jump_label(const fragment& fragment)
{
    return front_jump_label(fragment, 0);
}

offset_t assembler::back_jump_label(const fragment& fragment, offset_t offset)
{
    if (fragments.size() <= fragment.fragment_id) {
        throw invalid_fragment_error(fragment.fragment_id);
    }
    jump_labels.emplace_back(internal_jump_label{fragment.fragment_id, offset, false});
    return jump_labels.size() - 1;
}

offset_t assembler::back_jump_label(const fragment& fragment)
{
    if (fragments.size() <= fragment.fragment_id) {
        throw invalid_fragment_error(fragment.fragment_id);
    }
    return back_jump_label(fragment, fragments[fragment.fragment_id].size());
}


operation_list assembler::assemble() const
{
    auto offsets = fragment_offsets();

    /* Concatenate all fragments into a single script. While doing so, any
     * operation that uses the offset value will be patched to correctly
     * address the intended position in the intended fragment. */
    operation_list result;
    for (auto& fragment : fragments) {
        for (auto operation : fragment) {
            switch (operation.type) {
                /* All operations that use the offset value. */
                case operation_t::jump:
                case operation_t::cond:
                case operation_t::forward:
                case operation_t::forward_out_of_order:
                case operation_t::try_rlock:
                case operation_t::try_wlock:
                case operation_t::wait_for_votes:
                case operation_t::wait_for_yes_votes:
                case operation_t::wait_for_no_votes:
                case operation_t::wait_for_no_votes_es:
                case operation_t::wait_for_votes_end:
                    operation.offset = finalize_jump_label(offsets, operation.offset);
                    break;

                /* All operations that do not use the offset value. */
                case operation_t::end:
                case operation_t::execute_success:
                case operation_t::modify:
                case operation_t::rlock:
                case operation_t::wlock:
                case operation_t::unrlock:
                case operation_t::unwlock:
                case operation_t::vote_yes:
                case operation_t::vote_no:
                case operation_t::forceshutdown:
                    break;

            }
            result.emplace_back(operation);
        }
    }
    return result;
}

std::vector<offset_t> assembler::fragment_offsets() const
{
    std::vector<offset_t> result;
    offset_t sum = 0;
    for (auto& fragment : fragments) {
        result.emplace_back(sum);
        sum += fragment.size();
    }
    return result;
}

offset_t assembler::finalize_jump_label(const std::vector<offset_t>& offsets, offset_t jump_label) const
{
    if (jump_labels.size() <= jump_label) {
        throw invalid_jump_label_error(jump_label);
    }

    auto label = jump_labels[jump_label];

    auto& fragment = fragments[label.fragment_id];
    auto base = offsets[label.fragment_id];
    offset_t result = label.is_front_jump_label ? base + label.offset
                                                : base + fragment.size() - label.offset;

    if (!(base <= result && result < base + fragment.size())) {
        throw invalid_jump_label_error(jump_label);
    }

    return result;
}

/* Implementation of invalid_fragment_error. */

invalid_fragment_error::invalid_fragment_error(assembler::fragment_id_type fragment_id) : std::runtime_error(std::to_string(fragment_id)) {}


/* Implementation of invalid_jump_label_error. */

invalid_jump_label_error::invalid_jump_label_error(offset_t jump_label_id) : std::runtime_error(std::to_string(jump_label_id)) { }
