/**
 * Copyright (c) 2020 Jelle Hellings.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_TRANSACTION_ASSEMBLER_HPP
#define INCLUDE_TRANSACTION_ASSEMBLER_HPP
#include <deque>
#include <stdexcept>
#include <utility>
#include <vector>
#include "transaction_types.hpp"


namespace txn
{
    /**
     * The script assembler is designed to support the construction of scripts
     * with complex jump logic (e.g., due to jump, cond, try_rlocks, try_wlocks,
     * or forward).
     *
     * In plain scripts, the jump logic uses absolute offsets, which are hard to
     * determine before the script is constructed. This assembler replaces these
     * absolute offsets by jump-labels. To do so, this assembler builds a script
     * as a sequence of fragments F1; F2; F3; ...; Fn. If, during construction
     * of fragment Fi, one needs to jump to some operation in fragment Fj, one
     * simply generates a jump label (which provide fragment-relative offsets).
     * The assembler can generate jump labels in two ways:
     *
     * - a front-jump-label for offset x in fragment Fj will evaluate to the
     *   offset of the x-th operation in Fj (the first operation has offset 0).
     *
     * - a back-jump-label for offset x in fragment Fj will evaluate to the
     *   offset of the (Fj.size()-x)-th operation in Fj (the last operation
     *   has offset Fj.size()-1; hence, a reverse jump-label for offset x = 1
     *   will jump to that operation).
     *
     * Front-jump-labels can address static offsets in fragments that are built
     * from front-to-back; whereas back-jump-labels can address static offsets
     * in fragments that are built from back-to-front.
     *
     * When the assembler translates the sequence of fragments into a script,
     * these jump labels are translated into the correct absolute offsets in the
     * script.
     *
     * We notice that the assembler does not verify whether the created script
     * makes any sense. It will fail to assemble a script that used invalid jump
     * labels, however.
     */
    class assembler final
    {
    public:
        using fragment_operations_type = std::deque<operation>;
        using fragment_id_type = typename std::deque<fragment_operations_type>::size_type;


        /**
         * A fragment points to a list of operations, which can efficiently grow
         * via both ends, and has a unique fragment identifier.
         */
        struct fragment
        {
            fragment_operations_type& operations;
            fragment_id_type fragment_id;

            /**
             * Ability to treat this fragment as a pointer towards the list of
             * operations under construction.
             */
            fragment_operations_type* operator->()
            {
                return &operations;
            }
        };

        /**
         * Create a new fragment and return a pair (r, i); in which r is the
         * reference to the fragment and i is the identifier of the fragment.
         *
         * Each fragment is a double-ended queue of script operations.
         */
        fragment create_fragment();

        /**
         * Return a front-jump-label for offset in the specified fragment.
         *
         * Throws invalid_fragment_error if the specified fragment does not
         * exist.
         */
        offset_t front_jump_label(const fragment& fragment, offset_t offset);

        /**
         * Return a front-jump-label for the current start of the specified 
         * fragment.
         *
         * Throws invalid_fragment_error if the specified fragment does not
         * exist.
         */
        offset_t front_jump_label(const fragment& fragment);

        /**
         * Return a back-jump-label for offset in the specified fragment.
         *
         * Throws invalid_fragment_error if the specified fragment does not
         * exist.
         */
        offset_t back_jump_label(const fragment& fragment, offset_t offset);

        /**
         * Return a back-jump-label for the current start of the specified
         * fragment.
         *
         * Throws invalid_fragment_error if the specified fragment does not
         * exist.
         */
        offset_t back_jump_label(const fragment& fragment);

        /**
         * Generate the sequence of operations for this transaction script.
         *
         * Throws invalid_jump_label_error if any of the jump labels are invalid
         * (do not provide an offset within the specified fragment).
         */
        operation_list assemble() const;


        assembler() = default;
        assembler(const assembler&) = delete;
        assembler& operator=(const assembler&) = delete;
        assembler(assembler&&) = default;
        assembler& operator=(assembler&&) = default;


    private:
        /**
         * Internal representation of a jump label for the fragment_id-th
         * fragment. The jump-label is a front-jump-label to the specified
         * offset if the third value is true. Otherwise, the jump-label is a
         * back-jump-label.
         */
        struct internal_jump_label
        {
            fragment_id_type fragment_id;
            offset_t offset;
            bool is_front_jump_label;
        };


        /* The fragments in the assembled script. We use a deque, as we need
         * stable references to the individual fragments. */
        std::deque<fragment_operations_type> fragments;

        /* The created jump labels. */
        std::vector<internal_jump_label> jump_labels;


        /**
        * Given a list of fragments, return, for every fragment, the number of
        * operations in preceding fragments.
        */
        std::vector<offset_t> fragment_offsets() const;

        /**
         * Compute the final offset for the provided jump label based on the
         * information on the starting offset of individual fragments (as can be
         * computed via fragment_offsets).
         */
        offset_t finalize_jump_label(const std::vector<offset_t>& offsets, offset_t jump_label) const;
    };


    /**
     * Exception thrown by the assembler when encountering invalid fragment
     * identifiers.
     */
    struct invalid_fragment_error final : public std::runtime_error
    {
        invalid_fragment_error(assembler::fragment_id_type fragment_id);
    };

    /**
     * Exception thrown by the assembler when encountering invalid jump labels.
     */
    struct invalid_jump_label_error final : public std::runtime_error
    {
        invalid_jump_label_error(offset_t jump_label_id);
    };
}

#endif