/**
 * Copyright (c) 2020 Jelle Hellings.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "transaction_ostream.hpp"

#include <ostream>
#include <string>


template<class Container>
void print_helper(std::ostream& out, const Container& container, const std::string& indent)
{
    if (!container.empty()) {
        auto num_digits = std::to_string(container.size() - 1).size() + 2;

        typename Container::size_type i = 0;
        for (auto& entry : container) {
            
            auto line = std::to_string(i);

            out << indent << line << ':';
            for (auto d = line.size(); d < num_digits; ++d) {
                out << ' ';
            }

            out << entry << std::endl;
            ++i;
        }
    }
}


std::ostream& operator<<(std::ostream& out, const txn::transaction_atom& atom)
{
    switch (atom.type) {
        case txn::transaction_atom_t::check:
            out << "check(";
            break;
        case txn::transaction_atom_t::modify:
            out << "modify(";
            break;
    }
    return out << atom.account_id << ", " << atom.balance << ')';
}


std::ostream& operator<<(std::ostream& out, const txn::operation& operation)
{
    switch (operation.type) {
        case txn::operation_t::end:
            out << "end()";
            break;
        case txn::operation_t::execute_success:
            out << "execute_success()";
            break;
        case txn::operation_t::jump:
            out << "jump(" << operation.offset << ')';
            break;
        case txn::operation_t::cond:
            out << "cond(" << operation.account_id << ',' << operation.balance << ',' << operation.offset << ')';
            break;
        case txn::operation_t::modify:
            out << "modify(" << operation.account_id << ',' << operation.balance << ')';
            break;
        case txn::operation_t::forward:
            out << "forward(" << operation.shard_id << ',' << operation.offset << ')';
            break;
        case txn::operation_t::forward_out_of_order:
            out << "forward_out_of_order(" << operation.shard_id << ',' << operation.offset << ')';
            break;
        case txn::operation_t::rlock:
            out << "rlock(" << operation.account_id << ')';
            break;
        case txn::operation_t::wlock:
            out << "wlock(" << operation.account_id << ')';
            break;
        case txn::operation_t::try_rlock:
            out << "try_rlock(" << operation.account_id << ',' << operation.offset << ')';
            break;
        case txn::operation_t::try_wlock:
            out << "try_wlock(" << operation.account_id << ',' << operation.offset << ')';
            break;
        case txn::operation_t::unrlock:
            out << "unrlock(" << operation.account_id << ')';
            break;
        case txn::operation_t::unwlock:
            out << "unwlock(" << operation.account_id << ')';
            break;
        case txn::operation_t::wait_for_votes:
            out << "wait_for_votes(" << operation.num_votes << ',' << operation.shard_id << ',' << operation.offset << ')';
            break;
        case txn::operation_t::wait_for_yes_votes:
            out << "wait_for_yes_votes(" << operation.num_votes << ',' << operation.shard_id << ')';
            break;
        case txn::operation_t::wait_for_no_votes:
            out << "wait_for_no_votes(" << operation.num_votes << ',' << operation.shard_id << ')';
            break;
        case txn::operation_t::wait_for_no_votes_es:
            out << "wait_for_no_votes_es(" << operation.num_votes << ',' << operation.shard_id << ')';
            break;
        case txn::operation_t::wait_for_votes_end:
            out << "wait_for_votes_end(" << operation.num_votes << ')';
            break;
        case txn::operation_t::vote_yes:
            out << "vote_yes()";
            break;
        case txn::operation_t::vote_no:
            out << "vote_no()";
            break;
        case txn::operation_t::forceshutdown:
            out << "forceshutdown()";
            break;
     }
     return out;
}

std::ostream& operator<<(std::ostream& out, const txn::transaction& transaction)
{
    print_helper(out, transaction, "");
    return out;
}

std::ostream& operator<<(std::ostream& out, const txn::script& script)
{
    out << "Script " << script.script_id << " at " << script.initial_shard << std::endl;
    print_helper(out, script.operations, "    ");
    return out;
}
