/**
 * Copyright (c) 2020 Jelle Hellings.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "transaction_replica.hpp"


using namespace txn;


script_time::script_time(script_id_t script_id) : script_id(script_id), time_point(clock_type::now()) { }


/**
 * Helper function used to wake up script executions in lock_wait_queue. Write
 * the executions that can resume the provided wakeup_queue. If the
 * longest-waiting execution is waiting on a read lock, then wake up all
 * executions waiting for a read lock. Otherwise, wake up only the first
 * execution.
 */
void wakeup(std::deque<script_execution>& lock_wait_queue,
            std::deque<script_execution>& wakeup_queue)
{
    if (lock_wait_queue.empty()) {
        return;
    }

    auto& first = lock_wait_queue.front();
    if (first.script->operations[first.pc].type == operation_t::rlock) {
        /* Several transactions can hold the same read lock: we wake up each of
         * them to assure they all proceed execution quickly. */
        std::deque<script_execution> wlock_wait_queue;
        for (auto& value : lock_wait_queue) {
            if (value.script->operations[value.pc].type == operation_t::rlock) {
                wakeup_queue.emplace_back(value);
            }
            else {
                wlock_wait_queue.emplace_back(value);
            }
        }
        lock_wait_queue.swap(wlock_wait_queue);
    }
    else {
        /* Only a single transaction can hold the write lock. Hence, no need to
         * wake up more than a single transaction. */
        wakeup_queue.emplace_back(first);
        lock_wait_queue.pop_front();
    }
}


/* Implementation of account_state. */

bool details::account_state::try_read_lock(account& a)
{
    ++total_read_locks;

    /* Read lock: lock only when there is no concurrent write lock. */
    if (a.num_write_locked == 0) {
        ++a.num_read_locked;
        return true;
    }
    else {
        ++failed_read_locks;
        return false;
    }
}

bool details::account_state::try_write_lock(account& a)
{
    ++total_write_locks;

    /* Write lock: lock only when there is no concurrent lock. */
    if (a.num_read_locked == 0 && a.num_write_locked == 0) {
        ++a.num_write_locked;
        return true;
    }
    else {
        ++failed_write_locks;
        return false;
    }
}

bool details::account_state::read_lock(account& a, const script_execution& execution)
{
    auto lock = try_read_lock(a);
    if (!lock) {
        a.lock_wait_queue.emplace_back(execution);
    }
    return lock;
}

bool details::account_state::write_lock(account& a, const script_execution& execution)
{
    auto lock = try_write_lock(a);
    if (!lock) {
        a.lock_wait_queue.emplace_back(execution);
    }
    return lock;
}

bool details::account_state::read_unlock(account& a)
{
    if (a.num_read_locked == 0) {
        return false;
    }
    else {
        if (a.num_read_locked == 1) {
            wakeup(a.lock_wait_queue, wakeup_queue);
        }
        --a.num_read_locked;
        return true;
    }
}

bool details::account_state::write_unlock(account& a)
{
    if (a.num_write_locked == 0) {
        return false;
    }
    else {
        wakeup(a.lock_wait_queue, wakeup_queue);
        --a.num_write_locked;
        return true;
    }
}


/* Implementation of vote_waiting_state. */

details::vote_waiting_state::vote_wait_iterator details::vote_waiting_state::setup_voting(const script_execution& execution)
{
    ++total_vote_waits;
    auto result = vote_wait_queue.emplace(execution.script->script_id, vote_wait_entry{std::nullopt, 0, 0});
    result.first->second.execution = execution;
    return result.first;
}

details::vote_waiting_state::vote_wait_iterator details::vote_waiting_state::process_vote(txn::script_id_t script_id, bool vote)
{
    auto it = vote_wait_queue.emplace(script_id, vote_wait_entry{std::nullopt, 0, 0}).first;
    auto& vote_entry = it->second;
    if (vote) {
        ++total_yes_votes;
        ++vote_entry.num_yes_votes;
    }
    else {
        ++total_no_votes;
        ++vote_entry.num_no_votes;
    }
    return it;
}

std::optional<txn::details::vote_wait_entry> txn::details::vote_waiting_state::vote_wait_wakeup(vote_wait_iterator it)
{
    /* Voting is only done if this replica is waiting on the votes (execution is
     * set) and the number of votes in the current execution is equivalent to
     * the number of yes and no votes. */
    auto& vote_entry = it->second;
    if (vote_entry.execution.has_value()) {
        auto& execution = vote_entry.execution.value();
        auto operation = execution.script->operations[execution.pc];

        /* If voting is over, then process the wait. */
        if (operation.num_votes == vote_entry.num_yes_votes + vote_entry.num_no_votes) {
            vote_wait_entry result = vote_entry;
            vote_wait_queue.erase(it);
            return result;
        }
    }
    return std::nullopt;
}