/**
 * Copyright (c) 2020 Jelle Hellings.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_TRANSACTION_REPLICA_HPP
#define INCLUDE_TRANSACTION_REPLICA_HPP
#include <cstddef>
#include <deque>
#include <map>
#include <optional>
#include <utility>
#include "clock.hpp"
#include "transaction_system.hpp"
#include "utility_container.hpp"


namespace txn
{
    /**
     * To measure performance, replicas record pairs (script_id, current_time)
     * whenever a script begins, wakes up, or ends.
     */
    struct script_time
    {
        script_id_t script_id;
        clock_type::time_point time_point;


        /**
         * Construct the time point with the current time.
         */
        script_time(script_id_t script_id);
    };
 

    namespace details
    {
        /**
         * The state machines running at each replica keep track of detailed
         * measurements on their performance.
         */
        struct measurement
        {
            /* Identifiers of failed scripts. */
            std::deque<txn::script_id_t> failed_scripts = {};

            /* Total scripts that reach commit. */
            std::size_t executed_successful = 0;

            /* Performnce measurements for scripts: processing start times (upon
             * receipt from a client or another shard) and processing end times
             * (upon execution of an end-operation). */
            std::deque<script_time> script_begins = {};
            std::deque<script_time> script_ends = {};
        };


        /**
         * Performance statistics on read and write locks and succesful
         * executed scripts.
         */
        struct account_measurement
        {
            /* Performance measurements for scripts: wake up times for scripts
             * waiting on locks. */
            std::deque<script_time> script_wakeups = {};

            /* Condition failures (usually triggers abort code). */
            std::size_t condition_failures = 0;

            /* Forwards and out-of-order forwards */
            std::size_t total_forwards = 0;
            std::size_t total_forwards_out_of_order = 0;

            /* Information on locks. */
            std::size_t total_read_locks = 0;
            std::size_t total_write_locks = 0;
            std::size_t failed_read_locks = 0;
            std::size_t failed_write_locks = 0; 
        };


        /**
         * Performance statistics on vote-waits, yes votes, and no votes.
         */
        struct vote_measurement
        {
            std::size_t total_vote_waits = 0;
            std::size_t total_vote_waits_positive = 0;
            std::size_t total_yes_votes = 0;
            std::size_t total_no_votes = 0;
        };


        /**
         * Each replica holds a set of account states. Each state represents a
         * single account with a balance. Furthermore, each account can be read
         * and write locked independent of any other account, and the lock data
         * is also kept by the account state.
         */
        class account final
        {
        public:
            /* Balance of the account. */
            balance_t balance;


            /**
             * Construct a new account.
             */
            account(balance_t initial_balance) :
                    balance(initial_balance), num_read_locked(0), num_write_locked(0), lock_wait_queue() {}


            /* Data structures used to manage account locks. */
            std::size_t num_read_locked;
            std::size_t num_write_locked;

            /* Transactions waiting on an unlock operation. */
            std::deque<script_execution> lock_wait_queue;


            friend struct account_state;
        };


        /**
         * At each replica, two separate state machines are maintained: one
         * managing account information (and the associated account locks); the
         * other managing vote-waiting. Next is the internal state for the state
         * machine that manages account information (a set of accounts and the
         * associated lock structures).
         */
        struct account_state : public account_measurement
        {
            /* The accounts of this replica. */
            std::vector<account> accounts;

            /* The queue with all scripts that woke up due to an unlock. */
            std::deque<script_execution> wakeup_queue;

            /**
             * Non-blocking locks. Return true when the lock on the account can
             * be obtained, otherwise return false.
             */
            bool try_read_lock(account& a);
            bool try_write_lock(account& a);

            /**
             * Blocking locks. If the lock cannot be obtained, then the script
             * execution is placed on a wait list (from which it can wake up on
             * an unlock).
             */
            bool read_lock(account& a, const script_execution& execution);
            bool write_lock(account& a, const script_execution& execution);

            /**
             * Unlock the account. Will wake up script executions waiting for
             * the account if necessary.
             */
            bool read_unlock(account& a);
            bool write_unlock(account& a);


            /**
             * The internal account state is initialized with all account
             * balance data.
             */
            template<class Iterator>
            account_state(const Iterator begin, const Iterator end) :
                    account_measurement(), accounts(begin, end), wakeup_queue() { }
        };


        /**
         * Each vote-wait entry holds all data relevant for voting and waiting
         * on a single script (assuming unique script identifiers). Votes can be
         * collected before the waiting execution has been received. Hence,
         * the execution is optional.
         */
        struct vote_wait_entry
        {
            std::optional<script_execution> execution = std::nullopt;
            std::size_t num_yes_votes = 0;
            std::size_t num_no_votes = 0;
        };


        /**
         * The internal state for the state machine that manages vote-waiting.
         */
        struct vote_waiting_state : public vote_measurement
        {
            /* The vote-wait queue holds the data necessary to process votes
             * and vote-waiting. */
            std::map<txn::script_id_t, vote_wait_entry> vote_wait_queue;

            /* Pointer to a vote-wait entry. */
            using vote_wait_iterator = typename std::map<txn::script_id_t, vote_wait_entry>::iterator;


            /**
             * Setup a vote for the specified execution and return a pointer to
             * the vote entry.
             */
            vote_wait_iterator setup_voting(const script_execution& execution);

            /**
             * Process a vote. If voting is successful, then return a pointer to
             * the vote entry. Otherwise, return nothing.
             */
            vote_wait_iterator process_vote(txn::script_id_t script_id, bool vote);

             /**
             * Wake up the pointed-to entry from the vote_wait_queue if voting
             * completed.
             */
            std::optional<vote_wait_entry> vote_wait_wakeup(vote_wait_iterator it);
        };


        /**
         * Execution of a script can result in success, success_wait (when the
         * script could not obtain a blocking lock), failure (when an error in
         * the script is detected), or shutdown of the replica (due to executing
         * forceshutdown).
         */
        enum class execute_result
        {
            success,
            success_wait,
            failure,
            shutdown
        };
    }


    /**
     * A replica implements a basic state machine that executes externally-
     * provided scripts. Each replica implements a state machine that executes
     * scripts until execution of a forceshutdown-operation. All complex state
     * necessary for processing scripts is constructed before starting the state
     * machine. This assures that any measurements on the runtime of the replica
     * do not include unnecessary start-up times.
     *
     * Any script execution affecting account data needs to be done in a single
     * sequential order by each replica. Script execution that deals with
     * vote-waits and votes can be done in any arbitrary ordering in parallel to
     * executions affecting account data.
     *
     * To provide both, we provide two state machines, one that deals only with
     * account data and one that deals with vote-waits. These state macines can
     * be run in parallel at a single replica.
     */


    /**
     * The state machine that deals with account data.
     */
    template<txn::replica_system System>
    class account_state_machine : public details::measurement, public details::account_state
    {
    public:
        /* The replica system and replica id types used by this machine. */
        using system_type = System;
        using replica_id_type = const typename System::replica_id_type;

        /* The replica system and the identifier of this machine. */
        system_type& system;
        const replica_id_type replica_id;


        /**
         * Construct the replica state by initializing accounts using the
         * range of initial account balances.
         */
        template<class Iterator>
        account_state_machine(system_type& system, replica_id_type replica_id,
                              const Iterator begin, const Iterator end) : details::measurement(),
                                                                          details::account_state(begin, end),
                                                                          system(system), replica_id(replica_id) { }


        /**
         * Execute a script and return the result. The program counter wil
         * be updated during execution. Hence, the caller can determine the last
         * operation of the execution that was executed.
         */
        using execute_result = details::execute_result;
        execute_result execute(script_execution& execution); 

        /**
         * Select the next script to execute.
         */
        script_execution next_execution();
    };


    /**
     * The state machine that deals with vote-waits.
     */
    template<txn::replica_system System>
    class vote_state_machine : public details::measurement, public details::vote_waiting_state
    {
    public:
        /* The replica system and replica id types used by this machine. */
        using system_type = System;
        using replica_id_type = const typename System::replica_id_type;

        /* The replica system and the identifier of this machine. */
        system_type& system;
        const replica_id_type replica_id;

        /**
         * Construct the replica state.
         */
        vote_state_machine(system_type& system, replica_id_type replica_id) : details::measurement(),
                                                                              details::vote_waiting_state(),
                                                                              system(system), replica_id(replica_id) { }


        /**
         * Execute a script and return the result. The program counter wil
         * be updated during execution. Hence, the caller can determine the last
         * operation of the execution that was executed.
         */
        using execute_result = details::execute_result;
        execute_result execute(script_execution& execution); 

        /**
         * Select the next script to execute.
         */
        script_execution next_execution();

        /**
         * Forward execution based on the provided vote-wait entry.
         */
        void forward_vote_entry(const std::optional<details::vote_wait_entry>& entry);
    };


    /**
     * Execute the main loop of a state machine.
     */
    template<class StateMachine>
    void main_loop(StateMachine& machine)
    {
        using details::execute_result;

        while (true) {
            auto execution = machine.next_execution();
            switch (machine.execute(execution)) {
                case execute_result::success:
                    machine.script_ends.emplace_back(execution.script->script_id);
                    machine.system.release_execution(machine.replica_id, execution);
                    break;
                case execute_result::success_wait:
                    break;
                case execute_result::failure:
                    machine.failed_scripts.emplace_back(execution.script->script_id);
                    machine.system.release_execution(machine.replica_id, execution);
                    break;
                case execute_result::shutdown:
                    return;
            }
        }
    }

}


template<txn::replica_system System>
txn::script_execution txn::account_state_machine<System>::next_execution()
{
    /* A transaction will be returned from wakeup_queue if available. */
    if (!wakeup_queue.empty()) {
        auto v = util::front_pop(wakeup_queue);

        /* Waiting scripts do not send system messages: register start times. */
        script_wakeups.emplace_back(v.script->script_id);
        return v;
    }

    /* Otherwise, ask the system for a new script (from a client or forwarded by
     * another shard) to execute. */
    else {
        auto v = system.receive_execution(replica_id);

        /* We do only register the start-time of non-system messages. */
        if (v.script->script_id != script::system_script_id) {
            script_begins.emplace_back(v.script->script_id);
        }
        return v;
    }
}


template<txn::replica_system System>
txn::details::execute_result txn::account_state_machine<System>::execute(script_execution& execution)
{
    auto& operations = execution.script->operations;
    auto& pc = execution.pc;
    while (true) {
        /* Check: script did not go out-of-bounds. */
        if (operations.size() <= pc) {
            return execute_result::failure;
        }

        auto& op = operations[pc];
        auto next_pc = pc + 1;
        switch (op.type) {
            case operation_t::end:
                return execute_result::success;

            case operation_t::execute_success:
                ++executed_successful;
                break;

            case operation_t::jump:
                next_pc = op.offset;
                break;

            case operation_t::cond:
            {
                auto v = system.local_account_id(replica_id, op.account_id);
                if (!v.has_value()) {
                    return execute_result::failure;
                }
                if (accounts[v.value()].balance < op.balance) {
                    ++condition_failures;
                    next_pc = op.offset;
                }
                break;
            }

            case operation_t::modify:
            {
                auto v = system.local_account_id(replica_id, op.account_id);
                if (!v.has_value()) {
                    return execute_result::failure;
                }
                accounts[v.value()].balance += op.balance;
                break;
            }

            case operation_t::forward:
                if (!system.forward_execution(replica_id, op.shard_id, script_execution{execution.script, op.offset})) {
                    return execute_result::failure;
                }
                ++total_forwards;
                break;

            case operation_t::forward_out_of_order:
                if (!system.forward_execution_out_of_order(replica_id, op.shard_id, script_execution{execution.script, op.offset})) {
                    return execute_result::failure;
                }
                ++total_forwards_out_of_order;
                break;

            case operation_t::rlock:
            {
                auto v = system.local_account_id(replica_id, op.account_id);
                if (!v.has_value()) {
                    return execute_result::failure;
                }
                if (!read_lock(accounts[v.value()], execution)) {
                    return execute_result::success_wait;
                }
                break;
            }

            case operation_t::wlock:
            {
                auto v = system.local_account_id(replica_id, op.account_id);
                if (!v.has_value()) {
                    return execute_result::failure;
                }
                if (!write_lock(accounts[v.value()], execution)) {
                    return execute_result::success_wait;
                }
                break;
            }

            case operation_t::try_rlock:
            {
                auto v = system.local_account_id(replica_id, op.account_id);
                if (!v.has_value()) {
                    return execute_result::failure;
                }
                if (!try_read_lock(accounts[v.value()])) {
                    next_pc = op.offset;
                }
                break;
            }

            case operation_t::try_wlock:
            {
                auto v = system.local_account_id(replica_id, op.account_id);
                if (!v.has_value()) {
                    return execute_result::failure;
                }
                if (!try_write_lock(accounts[v.value()])) {
                    next_pc = op.offset;
                }
                break;
            }

            case operation_t::unrlock:
            {
                auto v = system.local_account_id(replica_id, op.account_id);
                if (!v.has_value()|| !read_unlock(accounts[v.value()])) {
                    return execute_result::failure;
                }
                break;
            }

            case operation_t::unwlock:
            {
                auto v = system.local_account_id(replica_id, op.account_id);
                if (!v.has_value()|| !write_unlock(accounts[v.value()])) {
                    return execute_result::failure;
                }
                break;
            }

            case operation_t::wait_for_votes:
            case operation_t::wait_for_yes_votes:
            case operation_t::wait_for_no_votes:
            case operation_t::wait_for_no_votes_es:
            case operation_t::wait_for_votes_end:
                if (!system.forward_execution_out_of_order(replica_id, replica_id, execution)) {
                    return execute_result::failure;
                }
                return execute_result::success;

            case operation_t::forceshutdown:
                return execute_result::shutdown;

            default:
                return execute_result::failure;
        }
        pc = next_pc;
    }
}


template<txn::replica_system System>
txn::script_execution txn::vote_state_machine<System>::next_execution()
{
    auto v = system.receive_execution_out_of_order(replica_id);
    if (v.script->script_id != script::system_script_id) {
        script_begins.emplace_back(v.script->script_id);
    }
    return v;
}


template<txn::replica_system System>
txn::details::execute_result txn::vote_state_machine<System>::execute(script_execution& execution)
{
    auto& operations = execution.script->operations;
    auto& pc = execution.pc;
    while (true) {
        /* Check: script did not go out-of-bounds. */
        if (operations.size() <= pc) {
            return execute_result::failure;
        }

        auto& op = operations[pc];
        auto next_pc = pc + 1;
        switch (op.type) {
            case operation_t::wait_for_votes:
            case operation_t::wait_for_yes_votes:
            case operation_t::wait_for_no_votes:
            case operation_t::wait_for_no_votes_es:
            case operation_t::wait_for_votes_end:
                if (op.num_votes == 0) {
                    return execute_result::failure;
                }
                forward_vote_entry(vote_wait_wakeup(setup_voting(execution)));
                return execute_result::success_wait;

            case operation_t::vote_yes:
                forward_vote_entry(vote_wait_wakeup(process_vote(execution.script->script_id, true)));
                return execute_result::success;

            case operation_t::vote_no:
                forward_vote_entry(vote_wait_wakeup(process_vote(execution.script->script_id, false)));
                return execute_result::success;

            case operation_t::forceshutdown:
                return execute_result::shutdown;

            default:
                return execute_result::failure;
        }
        pc = next_pc;
    }
}


template<txn::replica_system System>
void txn::vote_state_machine<System>::forward_vote_entry(const std::optional<details::vote_wait_entry>& entry)
{
    if (entry.has_value()) {
        auto& vote_wait_entry = entry.value();
        auto& execution = vote_wait_entry.execution.value();
        auto operation = execution.script->operations[execution.pc];

        txn::offset_t next_offset = execution.pc + 1;
        switch (operation.type) {
            case operation_t::wait_for_votes:
                if (vote_wait_entry.num_no_votes == 0) {
                    system.forward_execution(replica_id, operation.shard_id, script_execution{execution.script, next_offset});
                    ++total_vote_waits_positive;
                }
                else {
                    system.forward_execution(replica_id, operation.shard_id, script_execution{execution.script, operation.offset});
                }
                break;

            case operation_t::wait_for_yes_votes:
                if (vote_wait_entry.num_no_votes == 0) {
                    ++total_vote_waits_positive;
                    system.forward_execution(replica_id, operation.shard_id, script_execution{execution.script, next_offset});
                }
                break;

            case operation_t::wait_for_no_votes:
                if (vote_wait_entry.num_no_votes != 0) {
                    system.forward_execution(replica_id, operation.shard_id, script_execution{execution.script, next_offset});
                }
                else {
                    ++total_vote_waits_positive;
                }
                break;


            case operation_t::wait_for_no_votes_es:
                if (vote_wait_entry.num_no_votes != 0) {
                    system.forward_execution(replica_id, operation.shard_id, script_execution{execution.script, next_offset});
                }
                else {
                    ++total_vote_waits_positive;
                    ++executed_successful;
                }
                break;

            case operation_t::wait_for_votes_end:
                if (vote_wait_entry.num_no_votes == 0) {
                    ++total_vote_waits_positive;
                }
                break;

            default:
                failed_scripts.emplace_back(execution.script->script_id);
        }

        system.release_execution(replica_id, execution);
    }
}

#endif