/**
 * Copyright (c) 2020, 2021 Jelle Hellings.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_TRANSACTION_SYSTEM_HPP
#define INCLUDE_TRANSACTION_SYSTEM_HPP
#include <concepts>
#include <optional>
#include "transaction_types.hpp"


namespace txn
{
    /**
     * A sharded system consists of shards. Each such shard consists of a set of
     * replicas that each hold a subset of the data. In our design, basic
     * transactions require no knowledge of the division of replicas, shards,
     * and accounts. Furthermore, the replicas are implemented by basic state
     * machine that makes minimal assumptions of the system it runs in.
     *
     * To operate replicas and translate basic transactions into scripts that
     * can be run in a sharded system, replica state machines and the script
     * translation services rely on a system class that provides the necessary
     * details on the outside world.
     */

    /**
     * First, the requirements on a system that supports the translation of
     * transactions into scripts.
     */
    template<class System>
    concept translate_system =
        /**
         * Transaction translators require the ability to obtain fresh script
         * identifiers for each script---sequence of operations---they produce.
         */
        requires(System system, const txn::operation_list& operations) {
            { system.new_script_id(operations) } -> std::same_as<txn::script_id_t>;
        } &&

        /** 
         * Transaction translators require the ability to map from account
         * identifiers to the shard that account is assigned to.
         */
        requires(const System system, txn::account_id_t account_id) {
            { system.account_to_shard(account_id) } -> std::same_as<std::optional<txn::shard_id_t>>;
        } &&

        /**
         * AHL support requires the ability to obtain the shard id of the 
          * reference committee.
         */
        requires(System system) {
            { system.reference_committee_ahl_id() } -> std::same_as<txn::shard_id_t>;
        };


    /**
     * Then, the requirements on a system that is used by the state machine of
     * replicas to interact with the rest of the sharded system.
     */
    template<class System>
    concept replica_system =
        /**
         * Has a type to represent replica identifiers. (In some sharded
         * systems, each shard is a replica. In others, shards consists of many
         * full replicas, each with their own identifier).
         */
        requires {
            typename System::replica_id_type;
        } &&

        /**
         * Can return a script execution the replica needs to execute next
         * (e.g., from clients or other shards).
         */
        requires(System s, typename System::replica_id_type replica_id) {
            { s.receive_execution(replica_id) } -> std::same_as<txn::script_execution>;
        } &&
        requires(System s, typename System::replica_id_type replica_id) {
            { s.receive_execution_out_of_order(replica_id) } -> std::same_as<txn::script_execution>;
        } &&


        /**
         * Can forward execution of a script to another shard. Will return true
         * if the execution is forwarded or false otherwise (e.g., when the
         * shard identifier is invalid). There is also an out-of-order variant
         * that sends scripts that can be executed in any order and independent
         * from any other steps at the receiving shard.
         */
        requires(System s, typename System::replica_id_type from_replica_id,
                 txn::shard_id_t to_shard_id, txn::script_execution& execution) {
            { s.forward_execution(from_replica_id, to_shard_id, execution) } -> std::same_as<bool>;
        } &&
        requires(System s, typename System::replica_id_type from_replica_id,
                 txn::shard_id_t to_shard_id, txn::script_execution& execution) {
            { s.forward_execution_out_of_order(from_replica_id, to_shard_id, execution) } -> std::same_as<bool>;
        } &&

        /**
         * Can return an optional that holds the local account identifier for
         * the specified global account identifier (if that account is held by
         * this replica).
         */
        requires(const System s, typename System::replica_id_type replica_id, txn::account_id_t global_account_id) {
            { s.local_account_id(replica_id, global_account_id) } -> std::same_as<std::optional<txn::account_id_t>>;
        } &&

        /**
         * Can be signalled that the replica finished execution of an script.
         */
        requires(System s, typename System::replica_id_type replica_id, txn::script_execution execution) {
            { s.release_execution(replica_id, execution) } -> std::same_as<void>;
        };
}

#endif