/**
 * Copyright (c) 2020, 2021 Jelle Hellings.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "transaction_translate.hpp"

#include <algorithm>
#include <optional>
#include "transaction_assembler.hpp"
#include "transaction_utility.hpp"


/**
 * To keep the translation process simple and maintainable, we have opted to not
 * include elaborate pruning steps to remove death code. This in particularly
 * affects abort code, which is generated whether it is used or not. The only
 * focus of the translation is on correctness and on minimizing consensus and
 * cluster-sending (forward and forward_out_of_order). Hence, we can certainly
 * generate smaller scripts. For practical execution performance, this does not
 * have any benefits, however.
 */

using namespace txn;


/* Implementation of no_op_transaction_error. */
no_op_transaction_error::no_op_transaction_error() : std::runtime_error("target script is empty (invalid)") {}


/**
 * Type of isolation level for the generated scripts.
 */
enum class isolation_t
{
    read_uncommitted,
    read_committed,
    serializable
};

/**
 * Type of lock for the generated scripts.
 */
enum class lock_t
{
    blocking,
    wait_free
};

/**
 * Type of safety provided in optimistic level-0-isolation scripts.
 */
enum class optimistic_t
{
    unsafe,
    safe
};

/**
 * Type of orchestration in vote-based scripts.
 */
enum class vote_t
{
    centralized,
    distributed
};

/**
 * Type of execution in vote-based scripts.
 */
enum class vote_execute_t
{
    optimistic,
    optimistic_safe,
    read_uncommitted_wf,
    read_committed_wf,
    serializable_wf
};


/* Forward declaration of helper predicates. */
template<vote_execute_t> bool is_voter(const pre_translate_entry&);
template<vote_execute_t> bool is_waiter(const pre_translate_entry&);

/* Forward declaration of helper functions that generate code fragments that
 * deal with typical script functionality. */
template<vote_t, vote_execute_t> void operations_cast_votes(assembler&, const pre_translate_entry&, const shard_id_t root_shard_id, const pre_translate_data&,
                                                            assembler::fragment&, assembler::fragment&, assembler::fragment&, assembler::fragment&);
void operations_check_atom(assembler&, const txn::details::account_balance_pair, assembler::fragment&, assembler::fragment&);
void operations_check_atoms(assembler&, const pre_translate_entry&, assembler::fragment&, assembler::fragment&);
template<isolation_t, lock_t> void operations_fast_unlock_lock_account(assembler&, txn::details::account_read_only_flag_pair, std::optional<txn::details::account_balance_pair>,
                                                                       assembler::fragment&, assembler::fragment&, assembler::fragment&, assembler::fragment&);
template<isolation_t, lock_t> void operations_fast_unlock_lock(assembler&, const pre_translate_entry&, assembler::fragment&, assembler::fragment&, assembler::fragment&, assembler::fragment&);
void operations_forward_end(assembler&, const pre_translate_entry&, assembler::fragment&);
void operations_modify_all(const pre_translate_entry&, assembler::fragment&);
template<optimistic_t> void operations_modify_optimistic_fragments(assembler&, const pre_translate_entry&, assembler::fragment&, assembler::fragment&, assembler::fragment&);
template<optimistic_t> void operations_modify_optimistic(assembler&, const pre_translate_entry&, assembler::fragment&, assembler::fragment&, assembler::fragment&);
template<vote_execute_t> void operations_vote_local_abort_wait(const pre_translate_entry&, std::size_t, assembler::fragment&);
template<vote_t> void operations_vote_global_decide(assembler&, const pre_translate_entry&, assembler::fragment&, assembler::fragment&, assembler::fragment&, assembler::fragment&);
template<vote_t, vote_execute_t, optimistic_t> void operations_vote_optimistic(assembler&, const pre_translate_entry&, const shard_id_t root_shard_id, const pre_translate_data&, std::size_t, bool&,
                                                                               assembler::fragment&, assembler::fragment&, assembler::fragment&, assembler::fragment&, assembler::fragment&, assembler::fragment&);
template<vote_t, vote_execute_t, isolation_t> void operations_vote_lock_wf(assembler&, const pre_translate_entry&, const shard_id_t root_shard_id, const pre_translate_data&, std::size_t, bool&,
                                                                           assembler::fragment&, assembler::fragment&, assembler::fragment&, assembler::fragment&, assembler::fragment&, assembler::fragment&);


/* Implementation of helper predicates. */

/**
 * Return true if the provided shard is a voter.
 */
template<vote_execute_t execute_type>
bool is_voter(const pre_translate_entry& shard)
{
    switch (execute_type) {
        case vote_execute_t::optimistic:
        case vote_execute_t::optimistic_safe:
            /* Only check atoms cause voting. */
            return !shard.check_atoms.empty();

        case vote_execute_t::read_uncommitted_wf:
        case vote_execute_t::read_committed_wf:
        case vote_execute_t::serializable_wf:
            /* Modify atoms will be voted on due to wait-free write lock. Check
             * always cause voting. */
            return true;
    }
}

/**
 * Return true if the provided shard is a waiter.
 */
template<vote_execute_t execute_type>
bool is_waiter(const pre_translate_entry& shard)
{
    switch (execute_type) {
        case vote_execute_t::optimistic:
        case vote_execute_t::optimistic_safe:
            /* Optimistic scripts only commit or abort if they have
             * modifications. */
            return !shard.modify_atoms.empty();

        case vote_execute_t::read_uncommitted_wf:
        case vote_execute_t::read_committed_wf:
            /* In read uncommitted and read committed scripts, all modifications
             * are performed and write locks are released. */
            return !shard.modify_atoms.empty();

        case vote_execute_t::serializable_wf:
            /* In serializable scripts read and write locks need to be released
             * on commit. */
            return true;
    }
}


/* Implementation of code generation helper functions. */

/**
 * Write the required yes votes of shard to shard_yes_votes and no votes to
 * shard_no_votes.
 */
template<vote_t vote_type, vote_execute_t execute_type>
void operations_cast_votes(assembler& assembler, const pre_translate_entry& shard,
                           shard_id_t root_shard_id, const pre_translate_data& data,
                           assembler::fragment& shard_yes_vote, assembler::fragment& shard_no_vote,
                           assembler::fragment& yes_vote, assembler::fragment& no_vote)
{
    switch (vote_type) {
        case vote_t::centralized: {
            /* Send only yes/no vote to the root. */
            shard_yes_vote->emplace_back(operation_forward_out_of_order(root_shard_id, assembler.front_jump_label(yes_vote)));
            shard_no_vote->emplace_back(operation_forward_out_of_order(root_shard_id, assembler.front_jump_label(no_vote)));
            break;
        }
        case vote_t::distributed:
            /* Send yes/no vote to all waiters root. */
            for (auto& other : data) {
                if (other.shard_id != shard.shard_id && is_waiter<execute_type>(other)) {
                    shard_yes_vote->emplace_back(operation_forward_out_of_order(other.shard_id, assembler.front_jump_label(yes_vote)));
                    shard_no_vote->emplace_back(operation_forward_out_of_order(other.shard_id, assembler.front_jump_label(no_vote)));
                }
            }
            break;
    }
}

/**
 * Write the check to main that, on failure, goes to abort. We assume that
 * local_abort grows from the front.
 */
void operations_check_atom(assembler& assembler, const txn::details::account_balance_pair check_atom, assembler::fragment& main, assembler::fragment& abort)
{
    main->emplace_back(operation_cond(check_atom.account_id, check_atom.balance, assembler.back_jump_label(abort)));
}

/**
 * Write the checks to main that, on failure, goes to abort. We assume that
 * abort grows from the front.
 */
void operations_check_atoms(assembler& assembler, const pre_translate_entry& shard, assembler::fragment& main, assembler::fragment& abort)
{
    for (auto& atom : shard.check_atoms) {
        operations_check_atom(assembler, atom, main, abort);
    }
}

/**
 * Write fast-unlock lock-and-check for account to main. Write the necessary
 * unlock code for shard commit and abort, and local abort.
 */
template<isolation_t isolation_level, lock_t lock_type>
void operations_fast_unlock_lock_account(assembler& assembler, txn::details::account_read_only_flag_pair account, std::optional<txn::details::account_balance_pair> account_check,
                                        assembler::fragment& main, assembler::fragment& shard_commit, assembler::fragment& shard_abort, assembler::fragment& local_abort)
{
    /* A write lock is treated the same by all three isolation levels. */
    if (!account.read_only_flag) {
        auto lock_op = (lock_type == lock_t::blocking) ? operation_wlock(account.account_id)
                                                       : operation_try_wlock(account.account_id, assembler.back_jump_label(local_abort));
        main->emplace_back(lock_op);
        shard_commit->emplace_front(operation_unwlock(account.account_id));
        shard_abort->emplace_front(operation_unwlock(account.account_id));
        local_abort->emplace_front(operation_unwlock(account.account_id));
        if (account_check.has_value()) {
            operations_check_atom(assembler, account_check.value(), main, local_abort);
        }
    }
    else {
        /* The three isolation levels treat read locks very differently. Note
         * that there must be a check atom, as this account is read from (and
         * not written to). */
        switch (isolation_level) {
            case isolation_t::read_uncommitted:
                /* We do not need read locks, only perform the check. */
                operations_check_atom(assembler, account_check.value(), main, local_abort);
                break;

            case isolation_t::read_committed: {
                /* We do need a read lock, which we unlock directly after the
                 * check atom. */
                auto lock_op = (lock_type == lock_t::blocking) ? operation_rlock(account.account_id)
                                                               : operation_try_rlock(account.account_id, assembler.back_jump_label(local_abort));
                main->emplace_back(lock_op);

                /* We only need to unlock this account via a local abort if this
                 * check atom fails, after this check atom we immediately
                 * release the lock. Hence, make a separate code fragment that
                 * only performs this unlock. After this unlock, we still need
                 * to perform local abort for all write locks. */
                auto dedicated_local_abort = assembler.create_fragment();
                dedicated_local_abort->emplace_back(operation_unrlock(account.account_id));
                dedicated_local_abort->emplace_back(operation_jump(assembler.back_jump_label(local_abort)));
                operations_check_atom(assembler, account_check.value(), main, dedicated_local_abort);
                main->emplace_back(operation_unrlock(account.account_id));
                break;
            }

            case isolation_t::serializable: {
                /* Read locks the same way as write locks. */
                auto lock_op = (lock_type == lock_t::blocking) ? operation_rlock(account.account_id)
                                                               : operation_try_rlock(account.account_id, assembler.back_jump_label(local_abort));
                main->emplace_back(lock_op);
                shard_commit->emplace_front(operation_unrlock(account.account_id));
                shard_abort->emplace_front(operation_unrlock(account.account_id));
                local_abort->emplace_front(operation_unrlock(account.account_id));
                operations_check_atom(assembler, account_check.value(), main, local_abort);
                break;
            }
        }
    }
 }

 /**
  * Write code to get locks, verify checks, and release locks via the global
  * commit and abort paths and via the local abort path.
  */
 template<isolation_t isolation_level, lock_t lock_type>
 void operations_fast_unlock_lock(assembler& assembler, const pre_translate_entry& shard,
                                 assembler::fragment& main, assembler::fragment& shard_commit, assembler::fragment& shard_abort, assembler::fragment& local_abort)
{
    auto check_it = shard.check_atoms.cbegin(), check_end = shard.check_atoms.cend();
    for (auto account : shard.accounts) {
        /* Obtain the check atom for this account, if any. */
        std::optional<txn::details::account_balance_pair> account_check;
        if (check_it != check_end && check_it->account_id == account.account_id) {
            account_check = *check_it;
            ++check_it;
        }

        /* Set up the fast-unlock lock for account. */
        operations_fast_unlock_lock_account<isolation_level, lock_type>(assembler, account, account_check, main, shard_commit, shard_abort, local_abort);
    }
}

/**
 * Forward execution of main to shard, then end execution. Anything written to
 * main directly after this end, will be executed by shard. We assume that
 * main grows from the back.
 */
void operations_forward_end(assembler& assembler, const pre_translate_entry& shard, assembler::fragment& main)
{
    main->emplace_back(operation_forward(shard.shard_id, assembler.front_jump_label(main, main->size() + 2)));
    main->emplace_back(operation_end());
}

/**
 * Write all modifications of shard to the current back of perform.
 */
void operations_modify_all(const pre_translate_entry& shard, assembler::fragment& perform)
{
    for (auto& atom : shard.modify_atoms) {
        perform->emplace_back(operation_modify(atom.account_id, atom.balance));
    }
}

/**
 * Write modifications (commit and abort).
 */
template<optimistic_t optimistic_type>
void operations_modify_optimistic_fragments(assembler& assembler, const pre_translate_entry& shard,
                                            assembler::fragment& main, assembler::fragment& shard_commit, assembler::fragment& shard_abort)
{
    switch (optimistic_type) {
        case optimistic_t::unsafe: {
            /* Unsafe: write all changes to main and write rollback code to
             * shard_abort. */
            for (auto& atom : shard.modify_atoms) {
                main->emplace_back(operation_modify(atom.account_id, atom.balance));
                shard_abort->emplace_back(operation_modify(atom.account_id, -atom.balance));
            }
            break;
        }

        case optimistic_t::safe: {
            /* Safe: write all safe changes to main, all unsafe changes to
             * shard_commit, and rollback code to shard-abort. */
            for (auto& atom : shard.modify_atoms) {
                if (atom.balance > 0) {
                    shard_commit->emplace_back(operation_modify(atom.account_id, atom.balance));
                }
                else {
                    main->emplace_back(operation_modify(atom.account_id, atom.balance));
                    shard_abort->emplace_back(operation_modify(atom.account_id, -atom.balance));
                }
            }
        }
    }
}

/**
 * Write all modify atoms to the vote-step, commit-step, and abort-step in
 * accordance to the safety requirements.
 */
template<optimistic_t optimistic_type>
void operations_modify_optimistic(assembler& assembler, const pre_translate_entry& shard,
                                  assembler::fragment& main, assembler::fragment& commit, assembler::fragment& abort)
{
    auto shard_commit = assembler.create_fragment();
    auto shard_abort = assembler.create_fragment();
    operations_modify_optimistic_fragments<optimistic_type>(assembler, shard, main, shard_commit, shard_abort);
    if (!shard_commit->empty()) {
        shard_commit->emplace_back(operation_end());
        commit->emplace_back(operation_forward(shard.shard_id, assembler.front_jump_label(shard_commit)));
    }
    if (!shard_abort->empty()) {
        shard_abort->emplace_back(operation_end());
        abort->emplace_front(operation_forward(shard.shard_id, assembler.front_jump_label(shard_abort)));
    }
}

/**
 * Write the global commit or abort decision, if centralized orchestration is
 * used.
 */
template<vote_t vote_type>
void operations_vote_global_decide(assembler& assembler, const pre_translate_entry& shard,
                                   assembler::fragment& global_commit_root, assembler::fragment& global_abort_root,
                                   assembler::fragment& yes_vote, assembler::fragment& no_vote)
{
    if (vote_type == vote_t::centralized) {
        global_commit_root->emplace_back(operation_forward_out_of_order(shard.shard_id, assembler.front_jump_label(yes_vote)));
        global_abort_root->emplace_front(operation_forward_out_of_order(shard.shard_id, assembler.front_jump_label(no_vote)));
    }
}

/**
 * Write the code to finish local aborts. If this shard is a waiter, then
 * include a terminating wait-vote term that waits for the specified number of
 * votes.
 */
template<vote_execute_t execute_type>
void operations_vote_local_abort_wait(const pre_translate_entry& shard, std::size_t wait_votes, assembler::fragment& shard_no_vote_wait)
{
    if (is_waiter<execute_type>(shard)) {
        shard_no_vote_wait->emplace_back(operation_wait_for_votes_end(wait_votes, shard.shard_id));
    }
    else {
        shard_no_vote_wait->emplace_back(operation_end());
    }
}


/**
 * Write the vote, commit, and abort code for non-root shards in any of the
 * optimistic voting scripts.
 */
template<vote_t vote_type, vote_execute_t execute_type, optimistic_t optimistic_type>
void operations_vote_optimistic(assembler& assembler, const pre_translate_entry& shard,
                                const shard_id_t root_shard_id, const pre_translate_data& data,
                                std::size_t num_voters, bool& commit_operation_execute_success,
                                assembler::fragment& waiters, assembler::fragment& voters,
                                assembler::fragment& global_commit_root, assembler::fragment& global_abort_root,
                                assembler::fragment& yes_vote, assembler::fragment& no_vote)
{
    /* Shards without checks will only perfom actions on commit. */
    if (shard.check_atoms.empty()) {
        auto shard_commit = assembler.create_fragment();
        operations_modify_all(shard, shard_commit);
        shard_commit->emplace_back(operation_end());

        switch (vote_type) {
            case vote_t::centralized:
                /* If this is centralized, then add us to the global commit. */
                global_commit_root->emplace_back(operation_forward(shard.shard_id, assembler.front_jump_label(shard_commit)));
                break;

            case vote_t::distributed:
                /* Add this to the list of waiters and add the necessary
                 * vote-waiting. */
                if (!commit_operation_execute_success) {
                    shard_commit->emplace_front(operation_execute_success());
                    commit_operation_execute_success = true;
                }
                shard_commit->emplace_front(operation_wait_for_yes_votes(num_voters - 1, shard.shard_id));
                waiters->emplace_back(operation_forward_out_of_order(shard.shard_id, assembler.front_jump_label(shard_commit)));
                break;
        }
    }

    /* All other shards vote. */
    else {
        /* Main will perform the check atoms. All yes-votes will be written to
         * shard_yes_vote and all no votes to shard_no_vote_wait. If this shard
         * decides commit and will have to wait for a global decision, then
         * set up waiting in shard_yes_wait, global commit in shard_yes_commit
         * and global abort in shard_yes_abort. If this shard decides abort and
         * will have to wait for a global decision, then cast no votes and
         * set up waiting (which will ignore all votes, we already aborted) in
         * shard_no_vote_wait. */
        auto shard_main = assembler.create_fragment();
        auto shard_yes_vote = assembler.create_fragment();
        auto shard_yes_wait = assembler.create_fragment();
        auto shard_yes_commit = assembler.create_fragment();
        auto shard_yes_abort = assembler.create_fragment();
        auto shard_no_vote_wait = assembler.create_fragment();

        /* Set up the votes in shard_yes_vote and shard-no_vote. */
        operations_cast_votes<vote_type, execute_type>(assembler, shard, root_shard_id, data, shard_yes_vote, shard_no_vote_wait, yes_vote, no_vote);

        /* Set up the necessary actions on local abort decisions. */
        auto wait_votes = (vote_type == vote_t::centralized) ? 1 : num_voters - 2;
        operations_vote_local_abort_wait<execute_type>(shard, wait_votes, shard_no_vote_wait);

        /* Decide commit (fall-through to shard_yes_vote) or abort (jump to
         * shard_no_vote_wait). */
        operations_check_atoms(assembler, shard, shard_main, shard_no_vote_wait);

        /* Add commit and abort steps, if any. */
        if (!is_waiter<execute_type>(shard)) {
            /* If there are no modifications, then we can simply stop. We do not
             * need to wait for votes. */
            shard_yes_wait->emplace_back(operation_end());
        }
        else {
            /* Set up commit and abort code. */
            operations_modify_optimistic_fragments<optimistic_type>(assembler, shard, shard_main, shard_yes_commit, shard_yes_abort);

            /* Only process commit votes if we have commit code and abort code
             * if we have abort votes. */
            if (shard_yes_commit->empty()) {
                if (!commit_operation_execute_success) {
                    shard_yes_wait->emplace_back(operation_wait_for_no_votes_es(wait_votes, shard.shard_id));
                    commit_operation_execute_success = true;
                }
                else {
                    shard_yes_wait->emplace_back(operation_wait_for_no_votes(wait_votes, shard.shard_id));
                }
            }
            else {
                if (!commit_operation_execute_success) {
                    shard_yes_commit->emplace_back(operation_execute_success());
                    commit_operation_execute_success = true;
                }
                if (shard_yes_abort->empty()) {
                    shard_yes_wait->emplace_back(operation_wait_for_yes_votes(wait_votes, shard.shard_id));
                }
                else {
                    shard_yes_wait->emplace_back(operation_wait_for_votes(wait_votes, shard.shard_id, assembler.front_jump_label(shard_yes_abort)));
                    shard_yes_commit->emplace_back(operation_end());
                }
            }
            shard_yes_abort->emplace_back(operation_end());

            /* Set up the global commit or abort vote. */
            operations_vote_global_decide<vote_type>(assembler, shard, global_commit_root, global_abort_root, yes_vote, no_vote);
        }

        voters->emplace_back(operation_forward(shard.shard_id, assembler.front_jump_label(shard_main)));
    }
}

/**
 * Write the vote, commit, and abort code for non-root shards in any of the
 * wait-free lock-based voting scripts.
 */
template<vote_t vote_type, vote_execute_t execute_type, isolation_t isolation_level>
void operations_vote_lock_wf(assembler& assembler, const pre_translate_entry& shard,
                             const shard_id_t root_shard_id, const pre_translate_data& data,
                             std::size_t num_voters, bool& commit_operation_execute_success,
                             assembler::fragment& waiters, assembler::fragment& voters,
                             assembler::fragment& global_commit_root, assembler::fragment& global_abort_root,
                             assembler::fragment& yes_vote, assembler::fragment& no_vote)
{
    /* The structure of each shard is the same as in operations_vote_optimistic
     * (shards with check atoms). */
    auto shard_main = assembler.create_fragment();
    auto shard_yes_vote = assembler.create_fragment();
    auto shard_yes_wait = assembler.create_fragment();
    auto shard_yes_commit = assembler.create_fragment();
    auto shard_yes_abort = assembler.create_fragment();
    auto shard_no_vote_wait = assembler.create_fragment();

    /* Set up the votes in shard_yes_vote and shard-no_vote. */
    operations_cast_votes<vote_type, execute_type>(assembler, shard, root_shard_id, data, shard_yes_vote, shard_no_vote_wait, yes_vote, no_vote);

    /* Set up the necessary actions on local abort decisions. */
    auto wait_votes = (vote_type == vote_t::centralized) ? 1 : num_voters - 2;
    operations_vote_local_abort_wait<execute_type>(shard, wait_votes, shard_no_vote_wait);

    /* Obtain all locks, write the necessary release code, and decide commit
     * or abort along the way. */
    operations_fast_unlock_lock<isolation_level, lock_t::wait_free>(assembler, shard, shard_main, shard_yes_commit, shard_yes_abort, shard_no_vote_wait);

    /* Add global commit and abort steps, if any. */
    if (!is_waiter<execute_type>(shard)) {
        shard_yes_wait->emplace_back(operation_end());
    }
    else {
        /* Make all changes on commit. */
        operations_modify_all(shard, shard_yes_commit);
        
        /* Only process commit votes if we have commit code and abort code if we
         * have abort votes. */
        if (shard_yes_commit->empty()) {
            shard_yes_wait->emplace_back(operation_wait_for_no_votes(wait_votes, shard.shard_id));
        }
        else {
            if (!commit_operation_execute_success) {
                shard_yes_commit->emplace_back(operation_execute_success());
                commit_operation_execute_success = true;
            }
            if (shard_yes_abort->empty()) {
                shard_yes_wait->emplace_back(operation_wait_for_yes_votes(wait_votes, shard.shard_id));
            }
            else {
                shard_yes_wait->emplace_back(operation_wait_for_votes(wait_votes, shard.shard_id, assembler.front_jump_label(shard_yes_abort)));
                shard_yes_commit->emplace_back(operation_end());
            }
        }
        shard_yes_abort->emplace_back(operation_end());

        /* Set up the global commit or abort vote. */
        operations_vote_global_decide<vote_type>(assembler, shard, global_commit_root, global_abort_root, yes_vote, no_vote);
    }

    voters->emplace_back(operation_forward(shard.shard_id, assembler.front_jump_label(shard_main)));
}


/* Implementations of the main internal translation functions. */

/* Internal implementation of all linear orchestrated optimistic scripts. */
template<optimistic_t optimistic_type>
std::pair<operation_list, shard_id_t> internal_linear_optimistic(const pre_translate_reordered_data& rdata)
{
    auto& data = rdata.data;
    if (data.empty()) {
        throw no_op_transaction_error();
    }

    assembler assembler;

    /* The main fragment contains the normal execution flow. The commit fragment
     * will contain forwards to all shards that have commit actions (fall
     * through of main for the last voting shard). The abort fragment will
     * contain forwards to all shards that have abort actions (such that
     * jumping to the i-th entry before the last will abort the first i
     * shards that have abort actions). */
    auto main = assembler.create_fragment();
    auto commit = assembler.create_fragment();
    auto abort = assembler.create_fragment();
    commit->emplace_back(operation_execute_success());
    abort->emplace_back(operation_end());

    /* Now, we add each shard to the main path. We assume that there are shards
     * with checks and that these shards come first. */
    for (auto it = data.cbegin(), end = data.cend(); it != end; ++it) {
        auto next = it + 1;
        auto& shard = *it;

        /* This shard only performs actions on commit. */
        if (shard.check_atoms.empty()) {
            auto shard_commit = assembler.create_fragment();
            operations_modify_all(shard, shard_commit);
            shard_commit->emplace_back(operation_end());
            commit->emplace_back(operation_forward(shard.shard_id, assembler.front_jump_label(shard_commit)));
        }

        /* This shard performs checks in the main path. */
        else {
            /* If this is not the first shard, then the previous shard needs to
             * forward execution to this shard. */
            if (shard.shard_id != data.front().shard_id) {
                operations_forward_end(assembler, shard, main);
            }

            /* Perform the checks. */
            operations_check_atoms(assembler, shard, main, abort);

            /* Perform the modifications. */
            bool inline_commit = next == end || next->check_atoms.empty();
            if (inline_commit) {
                operations_modify_all(shard, main);
            }
            else {
                operations_modify_optimistic<optimistic_type>(assembler, shard, main, commit, abort);
            }
        }
    }

    commit->emplace_back(operation_end());
    return {assembler.assemble(), data.front().shard_id};
}

/* Internal implementation of all linear orchestrated lock-based scripts. */
template<isolation_t isolation_level, lock_t lock_type>
std::pair<operation_list, shard_id_t> internal_linear_lock_script(const pre_translate_data& data)
{
    if (data.empty()) {
        throw no_op_transaction_error();
    }

    assembler assembler;

    /* The main fragment contains the normal execution flow. The commit fragment
     * will contain forwards to all shards that have commit actions (fall
     * through of main for the last voting shard). The abort fragment will
     * contain forwards to all shards that have abort actions (such that
     * jumping to the i-th entry before the last will abort the first i
     * shards that have abort actions). */
    auto main = assembler.create_fragment();
    auto commit = assembler.create_fragment();
    auto abort = assembler.create_fragment();
    commit->emplace_back(operation_execute_success());
    abort->emplace_back(operation_end());

    /* Now, we add each shard to the main path. */
    for (auto& shard : data) {
        /* If this is not the first shard, then the previous shard needs to
         * forward execution to this shard. */
        if (shard.shard_id != data.front().shard_id) {
            operations_forward_end(assembler, shard, main);
        }

        /* We want to abort-fast: we want to perform each check atom in shard as
         * soon as we have the necessary locks to be able to do so. In this way,
         * we do not over-acquire locks when we have to abort the transaction
         * (improving performance when there is high lock contention).
         *
         * To do so, we process the accounts ordered by shard.accounts. For each
         * account, we get the necessary locks and perform the necessary check
         * atoms, and provide the necessary abort steps to deal with the case
         * when a check atom does not hold. In this sense, a wait-free lock is
         * just a special-purpose check atom.
         *
         * When a check atom in this shard fails, we only want to unlock the
         * locks we already have in this shard, after which we abort all
         * previous shards. Hence, we write a separate local abort (that handles
         * aborts when a check atom in this shard fails) and a shard abort that
         * handles abort votes from future shards.
         *
         * Finally, we also need a shard commit to perform all commit actions
         * whenever the global commit vote arrives. */
        auto local_abort = assembler.create_fragment();
        if (abort->size() == 1) {
            local_abort->emplace_back(operation_end());
        }
        else {
            local_abort->emplace_back(operation_jump(assembler.back_jump_label(abort)));
        }

        auto shard_abort = assembler.create_fragment();
        shard_abort->emplace_back(operation_end());

        auto shard_commit = assembler.create_fragment();
        operations_modify_all(shard, shard_commit);

        operations_fast_unlock_lock<isolation_level, lock_type>(assembler, shard, main, shard_commit, shard_abort, local_abort);

        /* Add the shard commit to the commit list if it has content. */
        if (!shard_commit->empty()) {
            shard_commit->emplace_back(operation_end());

            /* If this is the last shard, then we can jump to the global commit
             * of this shard (without forwarding). */
            if (shard.shard_id == data.back().shard_id) {
                commit->emplace_back(operation_jump(assembler.front_jump_label(shard_commit)));
            }
            else {
                commit->emplace_back(operation_forward(shard.shard_id, assembler.front_jump_label(shard_commit)));
            }
        }

        /* Add the shard abort to the abort list if it has content. */
        if (shard_abort->size() > 1) {
            abort->emplace_front(operation_forward(shard.shard_id, assembler.front_jump_label(shard_abort)));
        }
        else {
            shard_abort->clear();
        }
    }

    commit->emplace_back(operation_end());
    return {assembler.assemble(), data.front().shard_id};
}

/* Internal implementation of all vote-based scripts. */
template<vote_t vote_type, vote_execute_t execute_type>
std::pair<operation_list, shard_id_t> internal_vote_script(const pre_translate_reordered_data& rdata)
{
    auto& data = rdata.data;
    if (data.empty()) {
        throw no_op_transaction_error();
    }

    assembler assembler;

    /* To be able to produce correct wait_for_vote operations, we first
     * determine the total number of voters. */
    std::size_t num_voters = std::count_if(data.cbegin(), data.cend(), is_voter<execute_type>);

    /* Transactions with a few vote steps do not benefit from centralized or
     * distributed orchestration. To simplify this code, these transactions are
     * handled by linear orchestration. */
    if (num_voters < 3) {
        switch (execute_type) {
            case vote_execute_t::optimistic:
                return details::helper_linear_optimistic_unsafe(rdata);
            case vote_execute_t::optimistic_safe:
                return details::helper_linear_optimistic_safe(rdata);

            /* We use wait-free locks that never cause deadlock. Hence, we can
             * use the reordered entries with the normal linear lock-based
             * scripts. */
            case vote_execute_t::read_uncommitted_wf:
                return details::helper_linear_read_uncommitted_wf(data);
            case vote_execute_t::read_committed_wf:
                return details::helper_linear_read_committed_wf(data);
            case vote_execute_t::serializable_wf:
                return details::helper_linear_serializable_wf(data);
        }
    }

    /* We use the first shard as the root shard. */
    auto& root_shard = data.front();

    /* This function translates to two types of orchestration: centralized and
     * distributed orchestration.
     *
     * In centralized orchestration, only the root will receive all commit (yes)
     * and abort (no) votes. The root collects these votes and when it only
     * receives yes-votes, it will send out a commit message to all relevant
     * shards. Otherwise, it will send out an abort message.
     *
     * If we do not use locks, then commit and abort messages come in two
     * flavors: shards that did not have side-effects during voting (e.g., only
     * modify atoms; or check atoms and only unsafe modify atoms) will be
     * instructed to perform their changes during commit (similar to the commit
     * setup in undo-safe scripts). All other shards (that had side-effects
     * during voting) will receive a yes or no vote from the root indicating
     * commit or abort, respectively. This allows these shards to set up the
     * appropriate wait-handler to rollback any side-effects that might have to
     * be rolled back (e.g., if the shard voted commit, it would have performed
     * all safe modify atoms, which need to be rolled back only when the final
     * decision is abort; but if the shard voted abort, then it would not have
     * performed any modify atoms, and nothing needs to be rolled back).
     *
     * If we use locks, all shards need to vote. Not all shards have commit or
     * abort code, however: voters without modifications that do not enforce
     * serializability will unlock their read locks during voting.
     *
     * In distributed orchestration, all votes are sent to all shards that need
     * to perform either commit or abort steps. Each shard then independently
     * processes these votes and either commits or aborts their contribution of
     * the transaction.
     *
     * First, the root will check whether it can commit to the transaction
     * itself, and only if it can it will proceed with involving other shards.
     * The necessary code to do so is placed in main. Furthermore, if locks are
     * used, the main fragment will obtain all relevant locks. 
     *
     * We will place a forward to each non-root shard that needs to cast a vote
     * in the voters fragment. If distributed orchestration is enabled, then we
     * place a forward to each non-root shard that needs to wait for votes in
     * the waiters fragment. If the root needs to wait for votes, then it will
     * setup a wait-vote in the wait fragment. The commit_root fragment will
     * hold all steps the root must take to process a commit decision (local
     * modify atoms and---if centralized orhectration is used---send commit
     * messages), while the unlock_root fragment unlocks all locks held at the
     * root if locks are used. Finally, the abort_root fragment holds all steps
     * the root must take to process an abort decision (rollback atoms, unlock
     * accounts if locks are used, and---if centraized orchestration is used---
     * send abort messages). */
    auto main = assembler.create_fragment();
    auto voters = assembler.create_fragment();
    auto waiters = assembler.create_fragment();
    auto wait = assembler.create_fragment();
    auto global_commit_root = assembler.create_fragment();
    auto global_abort_root = assembler.create_fragment();
    auto local_abort_root = assembler.create_fragment();
    global_abort_root->emplace_back(operation_end());
    local_abort_root->emplace_back(operation_end());

    /* The root will always perform commit steps in centralized orchestration.
     * In decentralized orchestration, we add operation_execute_success to the
     * first shard one encounters with a commit fragment.*/
    bool commit_operation_execute_success = false;
    if (vote_type == vote_t::centralized || is_waiter<execute_type>(root_shard)) {
        global_commit_root->emplace_back(operation_execute_success());
        commit_operation_execute_success = true;
    }

    /* We cast yes and no votes by forwarding execution (out of order) to
     * dedicated yes and no operations. We only include this if there are
     * voters besides the root. */
    auto no_vote = assembler.create_fragment();
    auto yes_vote = assembler.create_fragment();
    no_vote->emplace_back(operation_vote_no());
    yes_vote->emplace_back(operation_vote_yes());

    /* Check whether the root can commit to this transaction and set up all
     * root execution steps. */
    switch (execute_type) {
        case vote_execute_t::optimistic:
            operations_check_atoms(assembler, root_shard, main, local_abort_root);
            operations_modify_optimistic_fragments<optimistic_t::unsafe>(assembler, root_shard, main, global_commit_root, global_abort_root);
            break;
        case vote_execute_t::optimistic_safe:
            operations_check_atoms(assembler, root_shard, main, local_abort_root);
            operations_modify_optimistic_fragments<optimistic_t::safe>(assembler, root_shard, main, global_commit_root, global_abort_root);
            break;
        case vote_execute_t::read_uncommitted_wf:
            operations_fast_unlock_lock<isolation_t::read_uncommitted, lock_t::wait_free>(assembler, root_shard, main, global_commit_root, global_abort_root, local_abort_root);
            operations_modify_all(root_shard, global_commit_root);
            break;
        case vote_execute_t::read_committed_wf:
            operations_fast_unlock_lock<isolation_t::read_committed, lock_t::wait_free>(assembler, root_shard, main, global_commit_root, global_abort_root, local_abort_root);
            operations_modify_all(root_shard, global_commit_root);
            break;
        case vote_execute_t::serializable_wf:
            operations_fast_unlock_lock<isolation_t::serializable, lock_t::wait_free>(assembler, root_shard, main, global_commit_root, global_abort_root, local_abort_root);
            operations_modify_all(root_shard, global_commit_root);
            break;
    }

    /* Include processing steps for all other shards. */
    for (auto& shard : data) {
        /* Skip the root shard. */
        if (shard.shard_id == root_shard.shard_id) {
            continue;
        }

        switch (execute_type) {
            case vote_execute_t::optimistic:
                operations_vote_optimistic<vote_type, execute_type, optimistic_t::unsafe>(assembler, shard, root_shard.shard_id, data, num_voters, commit_operation_execute_success,
                                                                                          waiters, voters, global_commit_root, global_abort_root, yes_vote, no_vote);
                break;
            case vote_execute_t::optimistic_safe:
                operations_vote_optimistic<vote_type, execute_type, optimistic_t::safe>(assembler, shard, root_shard.shard_id, data, num_voters, commit_operation_execute_success,
                                                                                        waiters, voters, global_commit_root, global_abort_root, yes_vote, no_vote);

                break;
            case vote_execute_t::read_uncommitted_wf:
                operations_vote_lock_wf<vote_type, execute_type, isolation_t::read_uncommitted>(assembler, shard, root_shard.shard_id, data, num_voters, commit_operation_execute_success,
                                                                                                waiters, voters, global_commit_root, global_abort_root, yes_vote, no_vote);
                break;
            case vote_execute_t::read_committed_wf:
                operations_vote_lock_wf<vote_type, execute_type, isolation_t::read_committed>(assembler, shard, root_shard.shard_id, data, num_voters, commit_operation_execute_success,
                                                                                              waiters, voters, global_commit_root, global_abort_root, yes_vote, no_vote);
                break;
            case vote_execute_t::serializable_wf:
                operations_vote_lock_wf<vote_type, execute_type, isolation_t::serializable>(assembler, shard, root_shard.shard_id, data, num_voters, commit_operation_execute_success,
                                                                                            waiters, voters, global_commit_root, global_abort_root, yes_vote, no_vote);
                break;
        }
    }

    /* Set up vote-waiting in the root shard. */
    if (vote_type == vote_t::centralized || is_waiter<execute_type>(root_shard)) {
        if (global_commit_root->empty()) {
            wait->emplace_back(operation_wait_for_no_votes(num_voters - 1, root_shard.shard_id));
        }
        else {
            wait->emplace_back(operation_wait_for_votes(num_voters - 1, root_shard.shard_id, assembler.front_jump_label(global_abort_root)));
            global_commit_root->emplace_back(operation_end());
        }
    }

    return {assembler.assemble(), root_shard.shard_id};
}


/* Implementation of the translate helper functions. */

std::pair<operation_list, shard_id_t> details::helper_linear_optimistic_unsafe(const pre_translate_reordered_data& rdata)
{
    return internal_linear_optimistic<optimistic_t::unsafe>(rdata);
}

std::pair<operation_list, shard_id_t> details::helper_linear_optimistic_safe(const pre_translate_reordered_data& rdata)
{
    return internal_linear_optimistic<optimistic_t::safe>(rdata);
}

std::pair<operation_list, shard_id_t> details::helper_linear_read_uncommitted(const pre_translate_data& data)
{
    return internal_linear_lock_script<isolation_t::read_uncommitted, lock_t::blocking>(data);
}

std::pair<operation_list, shard_id_t> details::helper_linear_read_uncommitted_wf(const pre_translate_data& data)
{
    return internal_linear_lock_script<isolation_t::read_uncommitted, lock_t::wait_free>(data);
}

std::pair<operation_list, shard_id_t> details::helper_linear_read_committed(const pre_translate_data& data)
{
    return internal_linear_lock_script<isolation_t::read_committed, lock_t::blocking>(data);
}

std::pair<operation_list, shard_id_t> details::helper_linear_read_committed_wf(const pre_translate_data& data)
{
    return internal_linear_lock_script<isolation_t::read_committed, lock_t::wait_free>(data);
}

std::pair<operation_list, shard_id_t> details::helper_linear_serializable(const pre_translate_data& data)
{
    return internal_linear_lock_script<isolation_t::serializable, lock_t::blocking>(data);
}

std::pair<operation_list, shard_id_t> details::helper_linear_serializable_wf(const pre_translate_data& data)
{
    return internal_linear_lock_script<isolation_t::serializable, lock_t::wait_free>(data);
}

std::pair<operation_list, shard_id_t> details::helper_centralized_optimistic_unsafe(const pre_translate_reordered_data& rdata)
{
    return internal_vote_script<vote_t::centralized, vote_execute_t::optimistic>(rdata);
}

std::pair<operation_list, shard_id_t> details::helper_centralized_optimistic_safe(const pre_translate_reordered_data& rdata)
{
    return internal_vote_script<vote_t::centralized, vote_execute_t::optimistic_safe>(rdata);
}

std::pair<operation_list, shard_id_t> details::helper_centralized_read_uncommitted_wf(const pre_translate_reordered_data& rdata)
{
    return internal_vote_script<vote_t::centralized, vote_execute_t::read_uncommitted_wf>(rdata);
}

std::pair<operation_list, shard_id_t> details::helper_centralized_read_committed_wf(const pre_translate_reordered_data& rdata)
{
    return internal_vote_script<vote_t::centralized, vote_execute_t::read_committed_wf>(rdata);
}

std::pair<operation_list, shard_id_t> details::helper_centralized_serializable_wf(const pre_translate_reordered_data& rdata)
{
    return internal_vote_script<vote_t::centralized, vote_execute_t::serializable_wf>(rdata);
}

std::pair<operation_list, shard_id_t> details::helper_distributed_optimistic_unsafe(const pre_translate_reordered_data& rdata)
{
    return internal_vote_script<vote_t::distributed, vote_execute_t::optimistic>(rdata);
}

std::pair<operation_list, shard_id_t> details::helper_distributed_optimistic_safe(const pre_translate_reordered_data& rdata)
{
    return internal_vote_script<vote_t::distributed, vote_execute_t::optimistic_safe>(rdata);
}

std::pair<operation_list, shard_id_t> details::helper_distributed_read_uncommitted_wf(const pre_translate_reordered_data& rdata)
{
    return internal_vote_script<vote_t::distributed, vote_execute_t::read_uncommitted_wf>(rdata);
}

std::pair<operation_list, shard_id_t> details::helper_distributed_read_committed_wf(const pre_translate_reordered_data& rdata)
{
    return internal_vote_script<vote_t::distributed, vote_execute_t::read_committed_wf>(rdata);
}

std::pair<operation_list, shard_id_t> details::helper_distributed_serializable_wf(const pre_translate_reordered_data& rdata)
{
    return internal_vote_script<vote_t::distributed, vote_execute_t::serializable_wf>(rdata);
}


/* The generator for AHL transactions. */

std::pair<operation_list, shard_id_t> details::helper_ahl(const pre_translate_reordered_data& rdata,
                                                          const shard_id_t reference_committee_id)
{
    /**
     * AHL scripts is based on internal_vote_scripts<vote_t::centralized, 
     * vote_execute_t::serializable_wf> with the following modifications:
     * - Only single-shard transactions are handled by linear orchestration.
     * - The root is always a dedicated reference shard (that does not manage
     *   accounts).
     *
     * We refer to the comments in internal_vote_scripts for details on the
     * structure of this code.
     */
    auto& data = rdata.data;
    if (data.empty()) {
        throw no_op_transaction_error();
    }

    assembler assembler;

    /* Determine the number of involved shards. If it is a single-shard
     * transaction, then use linear orchestration (which does a single-shard
     * consensus step. */
    std::size_t num_voters = std::count_if(data.cbegin(), data.cend(), is_voter<vote_execute_t::serializable_wf>);
    if (num_voters == 1) {
        return details::helper_linear_serializable_wf(data);
    }

    /* In AHL, the root is a reference committee---an independent shard that is
     * not involved in execution---that does not perform any execution steps
     * itself. Hence, we directly initiate the voting process. */
    auto voters = assembler.create_fragment();
    auto waiters = assembler.create_fragment();
    auto wait = assembler.create_fragment();
    auto global_commit_root = assembler.create_fragment();
    auto global_abort_root = assembler.create_fragment();
    auto local_abort_root = assembler.create_fragment();
    global_abort_root->emplace_back(operation_end());
    local_abort_root->emplace_back(operation_end());

    /* The reference committee can register successful execution. */
    global_commit_root->emplace_back(operation_execute_success());
    bool commit_operation_execute_success = true;

    /* Cast yes and no votes. */
    auto no_vote = assembler.create_fragment();
    auto yes_vote = assembler.create_fragment();
    no_vote->emplace_back(operation_vote_no());
    yes_vote->emplace_back(operation_vote_yes());

    /* Include processing steps for all shards. */
    for (auto& shard : data) {
            operations_vote_lock_wf<vote_t::centralized, vote_execute_t::serializable_wf, isolation_t::serializable>(
                    assembler, shard, reference_committee_id, data, num_voters, commit_operation_execute_success,
                    waiters, voters, global_commit_root, global_abort_root, yes_vote, no_vote);
    }

    /* Set up vote-waiting in the reference committee. Note that the reference
     * committee does not vote. Hence, it has to wait for all votes. */
    if (global_commit_root->empty()) {
        wait->emplace_back(operation_wait_for_no_votes(num_voters, reference_committee_id));
    }
    else {
        wait->emplace_back(operation_wait_for_votes(num_voters, reference_committee_id, assembler.front_jump_label(global_abort_root)));
        global_commit_root->emplace_back(operation_end());
    }

    return {assembler.assemble(), reference_committee_id};
}