/**
 * Copyright (c) 2020, 2021 Jelle Hellings.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_TRANSACTION_TRANSLATE_HPP
#define INCLUDE_TRANSACTION_TRANSLATE_HPP
#include "transaction_translate_pre.hpp"


namespace txn
{
    /**
     * We provide functionality to translate a transaction into a script. To
     * deal with system-specific issues (e.g., determining which accounts are on
     * which shards), the translators depend on a translate_system.
     *
     * The resultant scripts follow the orchestration-execution model: we
     * provide linear, centralized, and decentralized orchestration; and provide
     * optimistic execution, safe execution, two-phase locking, and wait-free
     * locking.
     *
     * The transaction translation process consists of three steps:
     *
     * (1) Pre-translation: construct a sequence of shard data entries, one for
     *     each shard affected by the transaction, such that each entry provided
     *     all the details for one such affected shards on the steps necessary
     *     to execute the transaction on a shard.
     *
     * (2) Reordering: for all scripts types (except linear orchestration with
     *     two-phase locking or wait-free locking), we then further optimize the
     *     pre-translation into an optimal ordering.
     *
     * (3) Translation: we translate the (reordered) pre-translation into a
     *     script (of the desired type).
     *
     * The pre-translate step is the same for all script types and the
     * reordering step is the same for all scripts that need it. Hence, if one
     * needs to create several script types for system s out of a transaction t,
     * then the most efficient way is to reuse the outcome of these steps. One
     * can do so as follows:
     *
     * // Step (1): create pre-tranlate data.
     * auto pd = pre_translate_transaction(t, s);
     *
     * // Step (2): create scripts with linear orchestration and two-phase
     * // locking or wait-free locking.
     * auto script_lru = translate_linear_read_uncommitted(pd, s);
     * auto script_lruwf = translate_linear_read_uncommitted_wf(pd, s);
     * auto script_lrc = translate_linear_read_committed(pd, s);
     * auto script_lrcwf = translate_linear_read_committed_wf(pd, s);
     * auto script_ls = translate_linear_serializable(pd, s);
     * auto script_lswf = translate_linear_serializable_wf(pd, s);
     *
     * // Step (3): Reordering.
     * auto pdr = pre_translate_reordering(std::move(pd));
     *
     * // Step (4): Create all other scripts.
     * auto script_lou = translate_linear_optimistic_unsafe(pdr, s);
     * auto script_los = translate_linear_optimistic_safe(pdr, s);
     * auto script_cou = translate_centralized_optimistic_unsafe(pdr, s);
     * auto script_cos = translate_centralized_optimistic_safe(pdr, s);
     * auto script_cruwf = translate_centralized_read_uncommitted_wf(pdr, s);
     * auto script_crcwf = translate_centralized_read_committed_wf(pdr, s);
     * auto script_cswf = translate_centralized_serializable_wf(pdr, s);
     * auto script_dou = translate_distributed_optimistic_unsafe(pdr, s);
     * auto script_dos = translate_distributed_optimistic_safe(pdr, s);
     * auto script_douwf = translate_distributed_read_uncommitted_wf(pdr, s);
     * auto script_docwf = translate_distributed_read_committed_wf(pdr, s);
     * auto script_doswf = translate_distributed_serializable_wf(pdr, s);
     * auto script_ahl = translate_ahl(pdr, s);
     */

    /**
     * Each translation function will throw no_op_transaction_error() when the
     * final script will not perform any operations (e.g., all atoms optimized
     * away).
     */
    struct no_op_transaction_error : std::runtime_error
    {
        no_op_transaction_error();
    };


    namespace details
    {
        /* Internal translate helpers. */
        std::pair<operation_list, shard_id_t> helper_linear_optimistic_unsafe(const pre_translate_reordered_data& rdata);
        std::pair<operation_list, shard_id_t> helper_linear_optimistic_safe(const pre_translate_reordered_data& rdata);
        std::pair<operation_list, shard_id_t> helper_linear_read_uncommitted(const pre_translate_data& data);
        std::pair<operation_list, shard_id_t> helper_linear_read_uncommitted_wf(const pre_translate_data& data);
        std::pair<operation_list, shard_id_t> helper_linear_read_committed(const pre_translate_data& data);
        std::pair<operation_list, shard_id_t> helper_linear_read_committed_wf(const pre_translate_data& data);
        std::pair<operation_list, shard_id_t> helper_linear_serializable(const pre_translate_data& data);
        std::pair<operation_list, shard_id_t> helper_linear_serializable_wf(const pre_translate_data& data);
        std::pair<operation_list, shard_id_t> helper_centralized_optimistic_unsafe(const pre_translate_reordered_data& rdata);
        std::pair<operation_list, shard_id_t> helper_centralized_optimistic_safe(const pre_translate_reordered_data& rdata);
        std::pair<operation_list, shard_id_t> helper_centralized_read_uncommitted_wf(const pre_translate_reordered_data& rdata);
        std::pair<operation_list, shard_id_t> helper_centralized_read_committed_wf(const pre_translate_reordered_data& rdata);
        std::pair<operation_list, shard_id_t> helper_centralized_serializable_wf(const pre_translate_reordered_data& rdata);
        std::pair<operation_list, shard_id_t> helper_distributed_optimistic_unsafe(const pre_translate_reordered_data& rdata);
        std::pair<operation_list, shard_id_t> helper_distributed_optimistic_safe(const pre_translate_reordered_data& rdata);
        std::pair<operation_list, shard_id_t> helper_distributed_read_uncommitted_wf(const pre_translate_reordered_data& rdata);
        std::pair<operation_list, shard_id_t> helper_distributed_read_committed_wf(const pre_translate_reordered_data& rdata);
        std::pair<operation_list, shard_id_t> helper_distributed_serializable_wf(const pre_translate_reordered_data& rdata);


        /* Support for AHL. */
        std::pair<operation_list, shard_id_t> helper_ahl(const pre_translate_reordered_data& rdata, const shard_id_t reference_committee_id);


        /**
         * Finish the script by adding the transaction identifier.
         */ 
        template<txn::translate_system System>
        script finish_script(std::pair<operation_list, shard_id_t>&& translate_output, System& system)
        {
            auto script_id = system.new_script_id(translate_output.first);
            return {script_id, translate_output.second, std::move(translate_output.first)};
        }
    }

   
    /**
     * The pre-translate-based transaction translation.
     */
    template<txn::translate_system System>
    script translate_linear_optimistic_unsafe(const pre_translate_reordered_data& rdata, System& system) {
        return details::finish_script(details::helper_linear_optimistic_unsafe(rdata), system);
    }

    template<txn::translate_system System>
    script translate_linear_optimistic_safe(const pre_translate_reordered_data& rdata, System& system) {
        return details::finish_script(details::helper_linear_optimistic_safe(rdata), system);
    }

    template<txn::translate_system System>
    script translate_linear_read_uncommitted(const pre_translate_data& data, System& system) {
        return details::finish_script(details::helper_linear_read_uncommitted(data), system);
    }

    template<txn::translate_system System>
    script translate_linear_read_uncommitted_wf(const pre_translate_data& data, System& system) {
        return details::finish_script(details::helper_linear_read_uncommitted_wf(data), system);
    }

    template<txn::translate_system System>
    script translate_linear_read_committed(const pre_translate_data& data, System& system) {
        return details::finish_script(details::helper_linear_read_committed(data), system);
    }

    template<txn::translate_system System>
    script translate_linear_read_committed_wf(const pre_translate_data& data, System& system) {
        return details::finish_script(details::helper_linear_read_committed_wf(data), system);
    }

    template<txn::translate_system System>
    script translate_linear_serializable(const pre_translate_data& data, System& system) {
        return details::finish_script(details::helper_linear_serializable(data), system);
    }

    template<txn::translate_system System>
    script translate_linear_serializable_wf(const pre_translate_data& data, System& system) {
        return details::finish_script(details::helper_linear_serializable_wf(data), system);
    }

    template<txn::translate_system System>
    script translate_centralized_optimistic_unsafe(const pre_translate_reordered_data& rdata, System& system) {
        return details::finish_script(details::helper_centralized_optimistic_unsafe(rdata), system);
    }

    template<txn::translate_system System>
    script translate_centralized_optimistic_safe(const pre_translate_reordered_data& rdata, System& system) {
        return details::finish_script(details::helper_centralized_optimistic_safe(rdata), system);
    }

    template<txn::translate_system System>
    script translate_centralized_read_uncommitted_wf(const pre_translate_reordered_data& rdata, System& system) {
        return details::finish_script(details::helper_centralized_read_uncommitted_wf(rdata), system);
    }

    template<txn::translate_system System>
    script translate_centralized_read_committed_wf(const pre_translate_reordered_data& rdata, System& system) {
        return details::finish_script(details::helper_centralized_read_committed_wf(rdata), system);
    }

    template<txn::translate_system System>
    script translate_centralized_serializable_wf(const pre_translate_reordered_data& rdata, System& system) {
        return details::finish_script(details::helper_centralized_serializable_wf(rdata), system);
    }

    template<txn::translate_system System>
    script translate_distributed_optimistic_unsafe(const pre_translate_reordered_data& rdata, System& system) {
        return details::finish_script(details::helper_distributed_optimistic_unsafe(rdata), system);
    }

    template<txn::translate_system System>
    script translate_distributed_optimistic_safe(const pre_translate_reordered_data& rdata, System& system) {
        return details::finish_script(details::helper_distributed_optimistic_safe(rdata), system);
    }

    template<txn::translate_system System>
    script translate_distributed_read_uncommitted_wf(const pre_translate_reordered_data& rdata, System& system) {
        return details::finish_script(details::helper_distributed_read_uncommitted_wf(rdata), system);
    }

    template<txn::translate_system System>
    script translate_distributed_read_committed_wf(const pre_translate_reordered_data& rdata, System& system) {
        return details::finish_script(details::helper_distributed_read_committed_wf(rdata), system);
    }

    template<txn::translate_system System>
    script translate_distributed_serializable_wf(const pre_translate_reordered_data& rdata, System& system) {
        return details::finish_script(details::helper_distributed_serializable_wf(rdata), system);
    }


    /**
     * Support for AHL.
     */
    template<txn::translate_system System>
    script translate_ahl(const pre_translate_reordered_data& rdata, System& system) {
        return details::finish_script(details::helper_ahl(rdata, system.reference_committee_ahl_id()), system);
    }
}

#endif