/**
 * Copyright (c) 2020 Jelle Hellings.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "transaction_translate_pre.hpp"

#include <algorithm>
#include <iterator>
#include <string>


using namespace txn;

translate_invalid_account_id::translate_invalid_account_id(account_id_t account_id) : std::runtime_error(std::to_string(account_id)) { }

/* Helpers for the implementation of details::pre_translate_transaction. */

/**
 * Return true if an atom would not perform any actions.
 */
bool is_no_op_atom(const transaction_atom& atom)
{
    return (atom.type == transaction_atom_t::modify && atom.balance == 0) ||
            (atom.type == transaction_atom_t::check && atom.balance == std::numeric_limits<balance_t>::min());
}

/**
 * Lexicographical ordering on shard_id, account_id, type.
 */
bool order_and_filter_less(const details::shard_atom_pair& lhs, const details::shard_atom_pair& rhs)
{
    return (lhs.shard_id < rhs.shard_id) ||
           ((lhs.shard_id == rhs.shard_id) && (lhs.atom.account_id < rhs.atom.account_id)) ||
           ((lhs.shard_id == rhs.shard_id) && (lhs.atom.account_id == rhs.atom.account_id) && (lhs.atom.type < rhs.atom.type));
}

/**
 * Order the annotated transaction on shard_id, account_id, type; merge all
 * atoms that can be merged; and delete all no-op atoms.
 */
void order_and_filter(details::annotated_transaction& transaction)
{
    std::sort(transaction.begin(), transaction.end(), order_and_filter_less);

    /* Simplify the list by merging consecutive atoms of the same type that
     * affect the same account. */
    auto wit = transaction.begin();
    for (auto rit = std::next(wit), end = transaction.end(); rit != end; ++rit) {
        /* If the new atom at rit cannot be merged with the atom at wit, then
         * add write rit as the next atom. */
        if ((rit->atom.account_id != wit->atom.account_id) || (rit->atom.type != wit->atom.type)) {
            /* Keep the last written atom only if it is not a no-op. */
            if (!is_no_op_atom(wit->atom)) {
                ++wit;
            }
            *wit = *rit;
        }
        else {
            /* The atoms ratom and watom refer to the same account and have the
             * same type: we can merge them. */
            if (wit->atom.type == transaction_atom_t::check) {
                wit->atom.balance = std::max(wit->atom.balance, rit->atom.balance);
            }
            else {
                wit->atom.balance += rit->atom.balance;
            }
        }
    }

    /* Keep the last written atom only if it is not a no-op. */
    if (!is_no_op_atom(wit->atom)) {
        ++wit;
    }

    /* Delete all entries after the last-written entry. */
    transaction.resize(std::distance(transaction.begin(), wit));
}



/* Implementation of details::pre_translate_transaction. */
pre_translate_data details::pre_translate_transaction(annotated_transaction& transaction)
{
    /* Order_and_filter does the heavy lifting. This function only reorganizes
     * all data into an easily-traversable structure. */
    order_and_filter(transaction);

    pre_translate_data result;
    for (auto pair : transaction) {
        /* The new atom affects the next shard. Make a new data entry. */
        if (result.empty() || result.back().shard_id != pair.shard_id) {
            result.emplace_back();
            result.back().shard_id = pair.shard_id;
        }

        /* The current data entry. */
        auto& entry = result.back();

        /* Add the atom to the next shard. */
        if (pair.atom.type == transaction_atom_t::check) {
            entry.check_atoms.emplace_back(details::account_balance_pair{pair.atom.account_id, pair.atom.balance});
        }
        else {
            entry.modify_atoms.emplace_back(details::account_balance_pair{pair.atom.account_id, pair.atom.balance});
            if (pair.atom.balance > 0) {
                ++entry.num_additions;
            }
        }

        /* Register the account for locks. */
        if (entry.accounts.empty() || entry.accounts.back().account_id != pair.atom.account_id) {
            entry.accounts.emplace_back(details::account_read_only_flag_pair{pair.atom.account_id, pair.atom.type == transaction_atom_t::check});
        }
        else if (pair.atom.type != transaction_atom_t::check) {
            entry.accounts.back().read_only_flag = false;
        }
    }

    return result;
}




/* Implementation for pre_translate_optimize_undo_order. */
pre_translate_reordered_data txn::pre_translate_reordering(pre_translate_data&& source)
{
    pre_translate_data no_abort_commit;
    pre_translate_data only_abort;
    pre_translate_data abort_commit;
    pre_translate_data modifications_only;

    for (auto&& entry : source) {
        /* No abort and no commit operations when there are no modifications. Do
         * these first. */
        if (entry.modify_atoms.empty()) {
            no_abort_commit.emplace_back(std::move(entry));
        }

        /* Can cause aborts due to checks. */
        else if (!entry.check_atoms.empty()) {
            /* If the code does only do safe modifications, then we only need
             * abort steps. Otherwise, we need bot abort and commit steps. */
            if (entry.num_additions == 0) {
                only_abort.emplace_back(std::move(entry));
            }
            else {
                abort_commit.emplace_back(std::move(entry));
            }
        }

        /* Cannot cause aborts and we do them last so that they also do not
         * need commit steps. */
        else {
            modifications_only.emplace_back(std::move(entry));
        }
    }

    pre_translate_reordered_data result{std::move(no_abort_commit)};
    result.data.insert(result.data.end(), std::make_move_iterator(only_abort.begin()), std::make_move_iterator(only_abort.end()));
    result.data.insert(result.data.end(), std::make_move_iterator(abort_commit.begin()), std::make_move_iterator(abort_commit.end()));
    result.data.insert(result.data.end(), std::make_move_iterator(modifications_only.begin()), std::make_move_iterator(modifications_only.end()));
    source.clear();
    return result;
}