/**
 * Copyright (c) 2020 Jelle Hellings.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_TRANSACTION_PRE_TRANSLATE_HPP
#define INCLUDE_TRANSACTION_PRE_TRANSLATE_HPP
#include <stdexcept>
#include <type_traits>
#include <vector>
#include "transaction_system.hpp"


namespace txn
{
    /**
     * Provides all pre-translation support for the translation of transactions
     * into scripts. We refer to transaction_translate.hpp for documentation on
     * how pre-translation fits within the overall translation process.
     */
    namespace details
    {
        /**
         * An (account identifier, balance) pair.
         */
        struct account_balance_pair final
        {
            account_id_t account_id;
            balance_t balance;
        };

        /**
         * An (account identifier, read-only flag) pair.
         */
        struct account_read_only_flag_pair final
        {
            account_id_t account_id;
            bool read_only_flag;
        };
    }

    /**
     * The base shard entry data necessary for the translation to a script of
     * the effects of a transaction onto a shard.
     */
    struct pre_translate_entry
    {
        /* The identifier of the shard that holds the accounts on which the
         * collected atoms should be applied. */
        shard_id_t shard_id;

        /* A pair (account_id, balance) for each check and each modify that
         * needs to be performed at this shard (at most a single check and
         * modify per account). Both lists will be ordered on increasing
         * account identifier. */
        std::vector<details::account_balance_pair> check_atoms;
        std::vector<details::account_balance_pair> modify_atoms;

        /* List of (account_id, flag) pairs with, for each account affected in
         * this shard, a single pair. The flag in a pair is true when only read-
         * only check atoms are specified for the account. The list will be
         * ordered on increasing account identifier. */
        std::vector<details::account_read_only_flag_pair> accounts;

        /* The number of accounts that will increase their balance in this shard
         * due to modify atoms with positive balances. */
        account_id_t num_additions;
    };


    /**
     * A list holding all pre_translate_shard_entry data derived from a
     * transaction.
     */
    using pre_translate_data = std::vector<pre_translate_entry>;


    /**
     * Exception thrown when pre-translating an account identifier that is not
     * recognized by the system.
     */
    struct translate_invalid_account_id : std::runtime_error
    {
        translate_invalid_account_id(account_id_t account_id);
    };


    namespace details
    {
        /**
         * A (shard_id, atom) pair that indicates the shard that holds the
         * account that the atom operates on (according to the translate_system
         * used).
         */
        struct shard_atom_pair final
        {
            shard_id_t shard_id;
            transaction_atom atom;
        };


        /* A shard-annotated transaction. */
        using annotated_transaction = std::vector<shard_atom_pair>;


        /**
         * Annotate the transaction by coupling each atom with the shard that
         * holds the account operated on by that atom.
         */
        template<txn::translate_system System>
        annotated_transaction shard_annotate_transaction(const transaction& transaction, const System& system)
        {
            annotated_transaction result;
            for (auto atom : transaction) {
                auto shard_id_opt = system.account_to_shard(atom.account_id);
                if (shard_id_opt.has_value()) {
                    result.emplace_back(shard_atom_pair{shard_id_opt.value(), atom});
                }
                else {
                    throw translate_invalid_account_id(atom.account_id);
                }
            }
            return result;
        }


        /** 
         * Pre-translate an annotated transaction.
         */
        pre_translate_data pre_translate_transaction(annotated_transaction& transaction);
    }


    /**
     * Pre-translate a transaction into a pre_translate_data structure. The
     * provided pre_translate_data structure is ordered on shard_id. Throws a
     * translate_invalid_account_id if any of the accounts in the transaction
     * are not part of the system
     */
    template<txn::translate_system System>
    pre_translate_data pre_translate_transaction(const transaction& transaction, const System& system)
    {
        auto annotated_transaction = details::shard_annotate_transaction(transaction, system);
        return details::pre_translate_transaction(annotated_transaction);
    }


    /**
     * During the translation, we require that the order in which shards are
     * processed is optimized in the sense that the amount of work during abort
     * operations is minimized and---at the same time---unsafe additions are
     * postponed as much as possible, e.g., if we have shards that do not have
     * check atoms, then we process these shards only after a commit decision is
     * already made and any modify atoms that add balance (and is unsafe) will
     * never rollback (altought optimistic execution never provides hard 
     * guarantees). To assure that un-optimized shard data entries are not
     * used interchangeably with optimized ones, we wrap optimized ones.
     *
     * We note that two-phase locking requires original non-reordered data.
     */
    struct pre_translate_reordered_data
    {
        pre_translate_data data;
    };


    /**
     * Perform the optimization reordering step. Will consume the input.
     */
    pre_translate_reordered_data pre_translate_reordering(pre_translate_data&& source);
}

#endif