/**
 * Copyright (c) 2020 Jelle Hellings.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_TRANSACTION_TYPES_HPP
#define INCLUDE_TRANSACTION_TYPES_HPP
#include <cstdint>
#include <limits>
#include <vector>


namespace txn
{
    /**
     * We provide high-level transactions and low-level scripts for performing
     * basic conditional account transformations in a sharded system, e.g.,
     * transferring funds from some source accounts to some target accounts
     * (which can be located on different shards) given that the sources all
     * have sufficient funds.
     *
     * On the one hand, the high-level transactions can express basic account
     * transformations without any details on how such transformations are
     * performed in a sharded system. On the other hand, the low-level scripts
     * provide full instructions to a sharded system on how to perform the
     * requested transformations (and on how to preserve data consistency
     * constraints).
     *
     * As such, transactions can be constructed without knowing the details of
     * the sharded system, whereas scripts are produced targeting a specific
     * sharded system configuration (that determines the number of shards and
     * the assignment of accounts to shards).
     */

    /* The type used for account identifiers. */
    using account_id_t = std::uint32_t;

    /* Type that holds balance-related values, e.g., the balance of an account
     * or a value to add to such a balance. */
    using balance_t = std::int32_t;


    /**
     * Each high-level transaction T is represented by a set of check and
     * modify atoms. Each check atom C in T verifies whether a given account
     * C.account_id has a balance of at least C.balance, whereas each modify
     * atom M in T adds M.balance to the balance of a given account
     * M.account_id (if M.balance is negative, then this modify constitutes a
     * subtraction of funds). With this interpretation of atoms, a high-level
     * transaction t will perform
     *
     *     ``If all checks in T hold, then perform all modifications in T.''
     */

    /* The type of atoms. */
    enum class transaction_atom_t : std::uint8_t
    {
        check,
        modify
    };

    /* A transaction atom. */
    struct transaction_atom final
    {
        transaction_atom_t type;
        account_id_t account_id;
        balance_t balance;
    };

    /* We store the set of atoms comprising a transaction in a plain vector. */
    using transaction = std::vector<transaction_atom>;


    /**
     * Scripts utilize a limited number of low-level operations that are
     * sufficient to implement transactions targeting a specific sharded system
     * configuration. All scripts follow the orchestration-execute model: while
     * processing a script, the shard executes each operation. Some of these
     * operations can forward execution of the script to other shards. The
     * low-level operations we provide are sufficient to implement transactions
     * that modify the data directly or that use locking to enforce data
     * consistency.
     */

    /* The type used for shard identifiers. */
    using shard_id_t = std::uint32_t;

    /* The type used for offsets in scripts. */
    using offset_t = std::uint16_t;


    /**
     * Operation code for script operations.
     */
    enum class operation_t : std::uint8_t
    {
        end,
        execute_success,
        jump,
        cond,
        modify,
        forward,
        forward_out_of_order,
        rlock,
        wlock,
        try_rlock,
        try_wlock,
        unrlock,
        unwlock,
        wait_for_votes,
        wait_for_yes_votes,
        wait_for_no_votes,
        wait_for_no_votes_es,
        wait_for_votes_end,
        vote_yes,
        vote_no,
        forceshutdown
    };


    /**
     * Representation of an operation in a script S. The interpretation of an
     * operation O in S is determined by the operation code S.type, and we
     * provide the following codes:
     *
     * - end(): terminate execution of the script by this replica. If a script
     *   performs multiple distinct steps at the same replica (e.g., due to the
     *   script returning to a replica via forwards), then the script will end
     *   multiple times at the same replica.
     *
     * - execute_success(): mark the execution of this script as successful
     *   (multiple calls will cause this script to be counted multiple times).
     *
     * - jump(offset): jump to the S.offset-th operation in the script
     *   (absolute).
     *
     * - cond(account_id, balance, offset): check whether the account
     *   S.account_id has a balance of at least S.balance. If this is the case,
     *   then continue execution with the next operation in the script.
     *   Otherwise, jump to the Sloffset-th operation in the script.
     *
     * - modify(account_id, balance): add S.balance to the account
     *   S.account_id. If S.balance is negative, then this modification
     *   constitutes a subtraction of funds.
     *
     * - forward(shard_id, offset): send a message to the shard S.shard_id
     *   instructing it to execute this script starting at the S.offset-th
     *   operation. This replica will continue execution with the next operation
     *   in the script.
     *
     * - forward_out_of_order(shard_id, offset): same as forward, except that
     *   the forwarded part of the script can be executed in any arbitrary
     *   order without affecting the outcome. Hence, we do not need to require
     *   ordering when processing.
     *
     * - rlock(account_id): read lock for S.account_id. Several scripts can
     *   obtain a read lock on the same account, but execution of this script
     *   will wait until all write locks on the account are released. This
     *   script must explicitly unlock the account later via unrlock.
     *
     * - wlock(account_id): write lock for S.account_id. Write locks are
     *   exclusive. Hence, execution of this script will wait until all locks on
     *   the account are released. This script must explicitly unlock the
     *   account later via unwlock.
     *
     * - try_rlock(account_id, offset): the non-blocking version of
     *   rlock(account_id). If the script obtains the read lock, then it
     *   continues execution with the next operation in the script. Otherwise,
     *   jump to the S.offset-th operation in the script.
     *
     * - try_wlock(account_id, offset): the non-blocking version of
     *   wlock(account_id).
     *
     * - unrlock(account_id): release a read lock on S.account_id. For any
     *   lock a script obtains, an explicit matching unlock operation must be
     *   executed (via every possible execution trace).
     *
     * - wait_for_votes(num_votes, shard_id, offset): wait for num_votes votes.
     *   If any of the votes is no, then forward execution to the specified
     *   shard at the specifie offset. Otherwise, forward execution to the
     *   specified shard at the next statement. Will require a forward
     *   (consensus) step for both outcomes. Can be processed out-of-order, will
     *   out-of-forward to itself if not out-of-order processing.
     *
     * - wait_for_yes_votes(num_votes, shard_id): wait for num_votes votes. If
     *   all votes are yes, then forward execution to the specified shard at the
     *   next statement. Otherwise, if some votes are no, then terminate
     *   execution of the script (see end()). Will only require a forward
     *   (consensus) step for yes outcomes. Can be processed out-of-order, will
     *   out-of-forward to itself if not out-of-order processing.
     *
     * - wait_for_no_votes(num_votes, shard_id): wait for num_votes votes. If
     *   all votes are no, then forward execution to the specified shard at the
     *   next statement. Otherwise, if all votes are yes, then terminate
     *   execution of the script (see end()). Will only require a forward
     *   (consensus) step for no outcomes. Can be processed out-of-order, will
     *   out-of-forward to itself if not out-of-order processing.
     *
     * - wait_for_no_votes_es(num_votes, shard_id): equivalent to
     *   wait_for_no_votes(num_votes, shard_id), except that it marks the
     *   execution of this script as successful if no no votes are received.
     *   This enables script types that do not have commit paths to mark
     *   succesful execution using their no-vote-based abort paths). See also
     *   execute_success().
     *
     * - wait_for_votes_end(num_votes, shard_id): wait for the specified number
     *   of votes, then terminate exection of the script (see end()). Will
     *   never require a forward (consensus) step for yes outcomes. Can be
     *   processed out-of-order, will out-of-forward to itself if not
     *   out-of-order processing.
     *
     * - vote_yes(): register a yes-vote for this script. Then terminate
     *   execution of the script (see end()). Must be processed out-of-order.
     *
     * - vote_no(): register a no-vote for this script. Then terminate execution
     *   of the script (see end()). Must be processed out-of-order.
     *
     * - unwlock(account_id): release a write lock on S.account_id.
     *
     * Furthermore, there is the forceshutdown() operation, which upon
     * processing at a replica terminates all operations at that replica. This
     * is used for gracefully terminating replicas.
     */
    struct operation final
    {
        operation_t type;
        offset_t offset;

        /* The operands account_id and balance; shard_id; and num_votes are
         * never used together. Hence, we take the union between these operands
         * to save space. We omit this optimization. */
        account_id_t account_id;
        balance_t balance;
        shard_id_t shard_id;
        shard_id_t num_votes;
    };


    /* The type used for script identifiers. */
    using script_id_t = std::uint64_t;

 
    /**
     * The core of a script is the list of operations it represents.
     */
    using operation_list = std::vector<operation>;


    /**
     * A script for a sharded system has a script identifier the system can use
     * to identify the script, the identifier of the shard that should start
     * execution of the script (at offset zero), and a list of operations.
     */
    struct script final
    {
        /* We reserve the last script identifier for system messages (replicas
         * will ignore these in their measurements). */
        constexpr static script_id_t system_script_id = std::numeric_limits<txn::script_id_t>::max();

        /* A system-used identifier for this script. */
        const script_id_t script_id;

        /* The initial shard for this script. */
        const shard_id_t initial_shard;

        /* The operations of the script. */
        const operation_list operations;
    };


    /**
     * During the execution of a script, one needs to maintains the execution
     * state. To present the execution state of a script S, one only need
     * access to the script and a program counter that point to the next
     * operation in S.
     */
    struct script_execution final
    {
        const txn::script* script;
        txn::offset_t pc;
    };
}

#endif