/**
 * Copyright (c) 2020 Jelle Hellings.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_TRANSACTION_UTILITY_HPP
#define INCLUDE_TRANSACTION_UTILITY_HPP
#include "transaction_types.hpp"


namespace txn
{
    /**
     * Create check atoms with the specified arguments.
     */

    constexpr transaction_atom atom_check(account_id_t account_id, balance_t balance)
    {
        return {transaction_atom_t::check, account_id, balance};
    }

    constexpr transaction_atom atom_modify(account_id_t account_id, balance_t balance)
    {
        return {transaction_atom_t::modify, account_id, balance};
    }


    /* Named helper values used in the construction of operations. */
    namespace details
    {
        constexpr offset_t no_offset = 0;
        constexpr account_id_t no_account_id = 0;
        constexpr balance_t no_balance = 0;
        constexpr shard_id_t no_shard_id = 0;
        constexpr shard_id_t no_num_votes = 0;
    }


    /**
     * Create operations with the specified arguments.
     */

    constexpr operation operation_end()
    {
        return {operation_t::end, details::no_offset, details::no_account_id,
                                  details::no_balance, details::no_shard_id, details::no_num_votes};
    }

    constexpr operation operation_execute_success()
    {
        return {operation_t::execute_success, details::no_offset, details::no_account_id, details::no_balance,
                                              details::no_shard_id, details::no_num_votes};
    }

    constexpr operation operation_jump(offset_t offset)
    {
        return {operation_t::jump, offset, details::no_account_id, details::no_balance,
                                   details::no_shard_id, details::no_num_votes};
    }

    constexpr operation operation_cond(account_id_t account_id, balance_t balance, offset_t offset)
    {
        return {operation_t::cond, offset, account_id, balance,
                                   details::no_shard_id, details::no_num_votes};
    }

    constexpr operation operation_modify(account_id_t account_id, balance_t balance)
    {
        return {operation_t::modify, details::no_offset, account_id, balance,
                                     details::no_shard_id, details::no_num_votes};
    }

    constexpr operation operation_forward(shard_id_t shard_id, offset_t offset)
    {
        return {operation_t::forward, offset, details::no_account_id, details::no_balance,
                                      shard_id, details::no_num_votes};
    }

    constexpr operation operation_forward_out_of_order(shard_id_t shard_id, offset_t offset)
    {
        return {operation_t::forward_out_of_order, offset, details::no_account_id, details::no_balance,
                                                  shard_id, details::no_num_votes};
    }

    constexpr operation operation_rlock(account_id_t account_id)
    {
        return {operation_t::rlock, details::no_offset, account_id, details::no_balance,
                                    details::no_shard_id, details::no_num_votes};
    }

    constexpr operation operation_wlock(account_id_t account_id)
    {
        return {operation_t::wlock, details::no_offset, account_id, details::no_balance,
                                    details::no_shard_id, details::no_num_votes};
    }

    constexpr operation operation_try_rlock(account_id_t account_id, offset_t offset)
    {
        return {operation_t::try_rlock, offset, account_id, details::no_balance,
                                        details::no_shard_id, details::no_num_votes};
    }

    constexpr operation operation_try_wlock(account_id_t account_id, offset_t offset)
    {
        return {operation_t::try_wlock, offset, account_id, details::no_balance,
                                        details::no_shard_id, details::no_num_votes};
    }

    constexpr operation operation_unrlock(account_id_t account_id)
    {
        return {operation_t::unrlock, details::no_offset, account_id, details::no_balance,
                                      details::no_shard_id, details::no_num_votes};
    }

    constexpr operation operation_unwlock(account_id_t account_id)
    {
        return {operation_t::unwlock, details::no_offset, account_id, details::no_balance,
                                      details::no_shard_id, details::no_num_votes};
    }

    constexpr operation operation_wait_for_votes(txn::shard_id_t num_votes, txn::shard_id_t shard_id, txn::offset_t offset)
    {
        return {operation_t::wait_for_votes, offset, details::no_account_id, details::no_balance,
                                             shard_id, num_votes};
    }

    constexpr operation operation_wait_for_yes_votes(txn::shard_id_t num_votes, txn::shard_id_t shard_id)
    {
        return {operation_t::wait_for_yes_votes, details::no_offset, details::no_account_id, details::no_balance,
                                                 shard_id, num_votes};
    }

    constexpr operation operation_wait_for_no_votes(txn::shard_id_t num_votes, txn::shard_id_t shard_id)
    {
        return {operation_t::wait_for_no_votes, details::no_offset, details::no_account_id, details::no_balance,
                                                shard_id, num_votes};
    }

    constexpr operation operation_wait_for_no_votes_es(txn::shard_id_t num_votes, txn::shard_id_t shard_id)
    {
        return {operation_t::wait_for_no_votes_es, details::no_offset, details::no_account_id, details::no_balance,
                                                   shard_id, num_votes};
    }

    constexpr operation operation_wait_for_votes_end(txn::shard_id_t num_votes, txn::shard_id_t shard_id)
    {
        return {operation_t::wait_for_votes_end, details::no_offset, details::no_account_id, details::no_balance,
                                                 shard_id, num_votes};
    }

    constexpr operation operation_vote_yes()
    {
        return {operation_t::vote_yes, details::no_offset, details::no_account_id, details::no_balance,
                                       details::no_shard_id, details::no_num_votes};
    }

    constexpr operation operation_vote_no()
    {
        return {operation_t::vote_no, details::no_offset, details::no_account_id, details::no_balance,
                                      details::no_shard_id, details::no_num_votes};
    }

    constexpr operation operation_forceshutdown()
    {
        return {operation_t::forceshutdown, details::no_offset, details::no_account_id, details::no_balance,
                                            details::no_shard_id, details::no_num_votes};
    }
}

#endif