/**
 * Copyright (c) 2020, 2021 Jelle Hellings.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_WORKLOAD_HPP
#define INCLUDE_WORKLOAD_HPP
#include <algorithm>
#include <type_traits>
#include <utility>
#include <random>
#include "transaction_system.hpp"
#include "transaction_translate.hpp"
#include "transaction_utility.hpp"


namespace workload
{
    /**
     * We provide flexible support to generate transaction workloads. The
     * support provided falls into the following categories:
     *
     * (1) The generation of specific types of transactional workloads.
     *
     * (2) The combination of several workloads into a single workload.
     *
     * (3) The translation of a high-level transactional workload into a set
     *     of low-level orchestration-execution scripts that can be run on a
     *     sharded system.
     *
     * (4) Generators to support the generation of workloads. We provide several
     *     number generators that can be used to generate, e.g., account numbers.
     */


    /**
     * A workload is simply a sequence of transactions. These can be translated
     * to scripts using a translate system to provide an actual executable
     * workload (for systems whose configuration match the translate system
     * used).
     */
    using transactions = std::vector<txn::transaction>;

    /**
     * An executable workload is a sequence of scripts.
     */
    using scripts = std::vector<txn::script>;


    /* (1) The generation of specific types of transactional workloads. */

    /**
     * Generate n transactions using the specified transaction generator. Each
     * time, the transaction generator is called with the arguments
     * transaction_args.
     */
    template<class TransactionGen, class... Args>
    transactions generate_transactions(std::size_t num_transactions, TransactionGen&& transaction_f, Args&&... transaction_args)
    {
        transactions result;
        while (result.size() < num_transactions) {
            result.emplace_back(transaction_f(std::forward<Args>(transaction_args)...));
        }
        return result;
    }


    /**
     * The dummy workload with factor n modifies the balance of n accounts by
     * incrementing them once. Consequently, the dummy workload provides an easy
     * workload to test maximum multi-shard script execution performance, as the
     * scripts do a minimum amount of work per shard. This workload is primarily
     * intended for debugging purposes
     *
     * Generate a single dummy transaction by drawing n = size_f() accounts 
     * via account_f().
     */
    struct dummy
    {
        template<class SizeGen, class AccountGen>
        txn::transaction operator()(SizeGen&& size_f, AccountGen&& account_f)
        {
            txn::transaction result;
            auto n = size_f();
            while (result.size() < n) {
                result.emplace_back(txn::atom_modify(account_f(), 1));
            }
            return result;
        }
    };


    /**
     * Construct a check-modify transaction that checks the balance of c =
     * check_f() accounts, removes some balance of i = input_f() accounts, and
     * add the removed balance evenly over o = output_f() accounts. Each of the
     * c + i + o accounts are drawn via account_f() calls and the c + i balances
     * are drawn via amount_f(). If safe is true, then the transaction will only
     * complete if all inputs have sufficient funds.
     */
    struct check_transfer
    {
        template<class CheckGen, class InputGen, class OutputGen, class AccountGen, class AmountGen>
        txn::transaction operator()(bool safe, CheckGen&& check_f, InputGen&& input_f, OutputGen&& output_f, AccountGen&& account_f, AmountGen&& amount_f)
        {
            txn::transaction result;
            std::vector<txn::account_id_t> accounts;

            /* We generate unique accounts for each check, input, and output,
             * this to assure that skewed account generation still affects
             * checks + inputs + outputs distinct accounts and not simply
             * results in a lot of operations on a single account. */
            auto checks = check_f();
            auto inputs = input_f();
            auto outputs =  output_f();
            while (accounts.size() != checks + inputs + outputs) {
                auto account_id = account_f();
                if (std::find(accounts.cbegin(), accounts.cend(), account_id) == accounts.cend()) {
                    accounts.emplace_back(account_id);
                }
            }

            /* To assure that a skewed account has equal probability of being
             * checked, being an input, or being an output, we shuffle the
             * generated accounts. */
            std::shuffle(accounts.begin(), accounts.end(), std::random_device());


            /* The checks. */
            for (std::size_t i = 0; i < checks; ++i) {
                result.emplace_back(txn::atom_check(accounts[i], amount_f()));
            }

            /* The inputs. */
            txn::balance_t total_input = 0;
            for (std::size_t i = 0; i < inputs; ++i) {
                txn::balance_t balance = amount_f();
                total_input += balance;
                result.emplace_back(txn::atom_modify(accounts[i + checks], -balance));

                if (safe) {
                    result.emplace_back(txn::atom_check(accounts[i + checks], balance));
                }
            }

            /* The outputs: smear the remainder r out over the first r
             * accounts. */
            for (std::size_t i = 0; i < outputs; ++i) {
                auto balance = (total_input / outputs);
                if (total_input % outputs < i) {
                    ++balance;
                }
                result.emplace_back(txn::atom_modify(accounts[i + checks + inputs], balance));
            }

            return result;
        }
    };


    /* (2) The combination of several workloads into a single workload. */

    /**
     * Combine the provided transactions in a single workload, e.g., to combine
     * different types of workloads together.
     */
    template<class... Args>
    transactions combine_transactions(Args&&... inputs)
    {
        transactions result;
        ((result.insert(result.end(), inputs.cbegin(), inputs.cend())), ...);
        std::shuffle(result.begin(), result.end(), std::random_device());
        return result;
    }


    /* (3) The translation of a high-level transactional workload. */

    /**
     * All protocols for which generate_scripts can generate output.
     */
    enum class protocol
    {
        linear_optimistic_unsafe,
        linear_optimistic_safe,
        linear_read_uncommitted,
        linear_read_uncommitted_wf,
        linear_read_committed,
        linear_read_committed_wf,
        linear_serializable,
        linear_serializable_wf,
        centralized_optimistic_unsafe,
        centralized_optimistic_safe,
        centralized_read_uncommitted_wf,
        centralized_read_committed_wf,
        centralized_serializable_wf,
        distributed_optimistic_unsafe,
        distributed_optimistic_safe,
        distributed_read_uncommitted_wf,
        distributed_read_committed_wf,
        distributed_serializable_wf,
        ahl
    };


    namespace details
    {

        /**
         * Compile-time helper to determine which protocols are selected or are
         * not selected.
         */
        template<protocol lhs, protocol rhs, protocol... rest>
        struct has_value
        {
            static constexpr bool value = (lhs == rhs) || has_value<lhs, rest...>::value;
        };

        template<protocol lhs, protocol rhs>
        struct has_value<lhs, rhs>
        {
            static constexpr bool value = lhs == rhs;
        };
    }

    /**
     * Enable all protocols.
     */
    struct protocols_all
    {
        static constexpr bool linear_optimistic_unsafe = true;
        static constexpr bool linear_optimistic_safe = true;
        static constexpr bool linear_read_uncommitted = true;
        static constexpr bool linear_read_uncommitted_wf = true;
        static constexpr bool linear_read_committed = true;
        static constexpr bool linear_read_committed_wf = true;
        static constexpr bool linear_serializable = true;
        static constexpr bool linear_serializable_wf = true;
        static constexpr bool centralized_optimistic_unsafe = true;
        static constexpr bool centralized_optimistic_safe = true;
        static constexpr bool centralized_read_uncommitted_wf = true;
        static constexpr bool centralized_read_committed_wf = true;
        static constexpr bool centralized_serializable_wf = true;
        static constexpr bool distributed_optimistic_unsafe = true;
        static constexpr bool distributed_optimistic_safe = true;
        static constexpr bool distributed_read_uncommitted_wf = true;
        static constexpr bool distributed_read_committed_wf = true;
        static constexpr bool distributed_serializable_wf = true;
        static constexpr bool ahl = true;
    };


    /**
     * Enable the protocols of interest by name. When provided to
     * generate_scripts, scripts will be produced for any protocol in Args.
     */
    template<protocol... Args>
    struct protocols_enable
    {
        static constexpr bool linear_optimistic_unsafe = details::has_value<protocol::linear_optimistic_unsafe, Args...>::value;
        static constexpr bool linear_optimistic_safe = details::has_value<protocol::linear_optimistic_safe, Args...>::value;
        static constexpr bool linear_read_uncommitted = details::has_value<protocol::linear_read_uncommitted, Args...>::value;
        static constexpr bool linear_read_uncommitted_wf = details::has_value<protocol::linear_read_uncommitted_wf, Args...>::value;
        static constexpr bool linear_read_committed = details::has_value<protocol::linear_read_committed, Args...>::value;
        static constexpr bool linear_read_committed_wf = details::has_value<protocol::linear_read_committed_wf, Args...>::value;
        static constexpr bool linear_serializable = details::has_value<protocol::linear_serializable, Args...>::value;
        static constexpr bool linear_serializable_wf = details::has_value<protocol::linear_serializable_wf, Args...>::value;
        static constexpr bool centralized_optimistic_unsafe = details::has_value<protocol::centralized_optimistic_unsafe, Args...>::value;
        static constexpr bool centralized_optimistic_safe = details::has_value<protocol::centralized_optimistic_safe, Args...>::value;
        static constexpr bool centralized_read_uncommitted_wf = details::has_value<protocol::centralized_read_uncommitted_wf, Args...>::value;
        static constexpr bool centralized_read_committed_wf = details::has_value<protocol::centralized_read_committed_wf, Args...>::value;
        static constexpr bool centralized_serializable_wf = details::has_value<protocol::centralized_serializable_wf, Args...>::value;
        static constexpr bool distributed_optimistic_unsafe = details::has_value<protocol::distributed_optimistic_unsafe, Args...>::value;
        static constexpr bool distributed_optimistic_safe = details::has_value<protocol::distributed_optimistic_safe, Args...>::value;
        static constexpr bool distributed_read_uncommitted_wf = details::has_value<protocol::distributed_read_uncommitted_wf, Args...>::value;
        static constexpr bool distributed_read_committed_wf = details::has_value<protocol::distributed_read_committed_wf, Args...>::value;
        static constexpr bool distributed_serializable_wf = details::has_value<protocol::distributed_serializable_wf, Args...>::value;
        static constexpr bool ahl = details::has_value<protocol::ahl, Args...>::value;
    };


    /**
     * Disable the protocols of interest by name. When provided to
     * generate_scripts, no scripts will be produced for any protocol in Args.
     */
    template<protocol... Args>
    struct protocols_disable
    {
        static constexpr bool linear_optimistic_unsafe = !details::has_value<protocol::linear_optimistic_unsafe, Args...>::value;
        static constexpr bool linear_optimistic_safe = !details::has_value<protocol::linear_optimistic_safe, Args...>::value;
        static constexpr bool linear_read_uncommitted = !details::has_value<protocol::linear_read_uncommitted, Args...>::value;
        static constexpr bool linear_read_uncommitted_wf = !details::has_value<protocol::linear_read_uncommitted_wf, Args...>::value;
        static constexpr bool linear_read_committed = !details::has_value<protocol::linear_read_committed, Args...>::value;
        static constexpr bool linear_read_committed_wf = !details::has_value<protocol::linear_read_committed_wf, Args...>::value;
        static constexpr bool linear_serializable = !details::has_value<protocol::linear_serializable, Args...>::value;
        static constexpr bool linear_serializable_wf = !details::has_value<protocol::linear_serializable_wf, Args...>::value;
        static constexpr bool centralized_optimistic_unsafe = !details::has_value<protocol::centralized_optimistic_unsafe, Args...>::value;
        static constexpr bool centralized_optimistic_safe = !details::has_value<protocol::centralized_optimistic_safe, Args...>::value;
        static constexpr bool centralized_read_uncommitted_wf = !details::has_value<protocol::centralized_read_uncommitted_wf, Args...>::value;
        static constexpr bool centralized_read_committed_wf = !details::has_value<protocol::centralized_read_committed_wf, Args...>::value;
        static constexpr bool centralized_serializable_wf = !details::has_value<protocol::centralized_serializable_wf, Args...>::value;
        static constexpr bool distributed_optimistic_unsafe = !details::has_value<protocol::distributed_optimistic_unsafe, Args...>::value;
        static constexpr bool distributed_optimistic_safe = !details::has_value<protocol::distributed_optimistic_safe, Args...>::value;
        static constexpr bool distributed_read_uncommitted_wf = !details::has_value<protocol::distributed_read_uncommitted_wf, Args...>::value;
        static constexpr bool distributed_read_committed_wf = !details::has_value<protocol::distributed_read_committed_wf, Args...>::value;
        static constexpr bool distributed_serializable_wf = !details::has_value<protocol::distributed_serializable_wf, Args...>::value;
        static constexpr bool ahl = !details::has_value<protocol::ahl, Args...>::value;
    };


    /**
     * Generate a workload from a set of transactions. Only generate the
     * selected script types.
     */
    template<txn::translate_system System, class Protocols>
    std::vector<scripts> generate_scripts(const transactions& transactions, System& system, const Protocols& p)
    {
        scripts lous, los, lru, lruwf, lrc, lrcwf, ls, lswf,
                cous, cos, cruwf, crcwf, cswf,
                dous, dos, druwf, drcwf, dswf,
                ahl;

        for (auto& t : transactions) {
            auto pd = txn::pre_translate_transaction(t, system);
            if (p.linear_read_uncommitted) lru.emplace_back(txn::translate_linear_read_uncommitted(pd, system)); 
            if (p.linear_read_uncommitted_wf) lruwf.emplace_back(txn::translate_linear_read_uncommitted_wf(pd, system));
            if (p.linear_read_committed) lrc.emplace_back(txn::translate_linear_read_committed(pd, system));
            if (p.linear_read_committed_wf) lrcwf.emplace_back(txn::translate_linear_read_committed_wf(pd, system));
            if (p.linear_serializable) ls.emplace_back(txn::translate_linear_serializable(pd, system));
            if (p.linear_serializable_wf) lswf.emplace_back(txn::translate_linear_serializable_wf(pd, system));

            if (p.linear_optimistic_unsafe || p.linear_optimistic_safe ||
                    p.centralized_optimistic_unsafe || p.centralized_optimistic_safe || p.centralized_read_uncommitted_wf || p.centralized_read_committed_wf || p.centralized_serializable_wf ||
                    p.distributed_optimistic_unsafe || p.distributed_optimistic_safe || p.distributed_read_uncommitted_wf || p.distributed_read_committed_wf || p.distributed_serializable_wf ||
                    p.ahl) {
                auto pdr = txn::pre_translate_reordering(std::move(pd));
                if (p.linear_optimistic_unsafe) lous.emplace_back(txn::translate_linear_optimistic_unsafe(pdr, system));
                if (p.linear_optimistic_safe) los.emplace_back(txn::translate_linear_optimistic_safe(pdr, system));
                if (p.centralized_optimistic_unsafe) cous.emplace_back(txn::translate_centralized_optimistic_unsafe(pdr, system));
                if (p.centralized_optimistic_safe) cos.emplace_back(txn::translate_centralized_optimistic_safe(pdr, system));
                if (p.centralized_read_uncommitted_wf) cruwf.emplace_back(txn::translate_centralized_read_uncommitted_wf(pdr, system));
                if (p.centralized_read_committed_wf) crcwf.emplace_back(txn::translate_centralized_read_committed_wf(pdr, system));
                if (p.centralized_serializable_wf) cswf.emplace_back(txn::translate_centralized_serializable_wf(pdr, system));
                if (p.distributed_optimistic_unsafe) dous.emplace_back(txn::translate_distributed_optimistic_unsafe(pdr, system));
                if (p.distributed_optimistic_safe) dos.emplace_back(txn::translate_distributed_optimistic_safe(pdr, system));
                if (p.distributed_read_uncommitted_wf) druwf.emplace_back(txn::translate_distributed_read_uncommitted_wf(pdr, system));
                if (p.distributed_read_committed_wf) drcwf.emplace_back(txn::translate_distributed_read_committed_wf(pdr, system));
                if (p.distributed_serializable_wf) dswf.emplace_back(txn::translate_distributed_serializable_wf(pdr, system));
                if (p.ahl) ahl.emplace_back(txn::translate_ahl(pdr, system));
            }
        }

        return {lous, los, lru, lruwf, lrc, lrcwf, ls, lswf,
                cous, cos, cruwf, crcwf, cswf,
                dous, dos, druwf, drcwf, dswf,
                ahl};
    }


    /* (4) Generators to support the generation of workloads. */

    /**
     * A generator that returns a constant value.
     */
    template<class Type>
    struct constant_generator
    {
        /* The type of the value. */
        using value_type = Type;

        /* The constant value. */
        const value_type value;

        /**
         * Return the constant value on every possible call.
         */
        template<class... Args>
        constexpr const value_type& operator()(Args&&... args) const
        {
            return value;
        }
    };

    /**
     * Return a constant generator that always returns the specified value.
     */
    template<class Type>
    constexpr constant_generator<Type> make_constant_generator(const Type& value)
    {
        return constant_generator<Type>{value};
    }


    /**
     * A generator that returns values in the range [min, max] (including min
     * and max). The i-th value returned is min + i with wrap around (we return
     * min again after returning max).
     */
    template<class Type>
    struct range_generator
    {
        /* The type of the value. */
        using value_type = Type;

        /* The minimum and maximum values. */
        const value_type min, max;

        /* The next-returned value. */
        value_type next;

        /**
         * Return the constant value on every possible call.
         */
        template<class... Args>
        constexpr const value_type operator()(Args&&... args)
        {
            if (next == max) {
                next = min;
                return max;
            }
            else {
                return next++;
            }
        }


        /**
         * Construct the range generator.
         */
        range_generator(const value_type& min, const value_type& max) : min(min), max(max), next(min) { }
    };

    /**
     * Return a range generator.
     */
    template<class Type>
    constexpr range_generator<Type> make_range_generator(const Type& min, const Type& max)
    {
        return range_generator<Type>(min, max);
    }


    /**
     * A generator that returns skewed values in the range [min, max] (including
     * min and max). To determine the likelihood of a values, we use a one-time
     * random shuffle on the range [min, max], resulting in list shuffled in
     * which shuffled[i] is the i-th likeliest value.
     *
     * The provided generator is used to determine the probability by which the
     * i-th likeliest value is chosen. E.g., if the generator produces indices
     * in the range [0, max - min + 1] with a geometric distribution, then the
     * i-th likeliest value in [min, max] has probability p * (1 - p)^i.
     */
    template<class Type, class Generator>
    struct skewed_generator
    {
        /* The type of the value. */
        using value_type = Type;

        /* The type of the generator. */
        using generator_type = Generator;

        /* The ordering of the values. */
        const std::vector<value_type> shuffled;

        /* The generator to choose incides in the range [0, max - min + 1]. Any
         * output outside of that range is ignored. */
        generator_type& generator;

        /**
         * Return the constant value on every possible call.
         */
        template<class... Args>
        constexpr const value_type operator()(Args&&... args)
        {
            auto i = generator();
            while (!(0 < i && i < shuffled.size())) {
                i = generator();
            }
            return shuffled[i];
        }


        /**
         * Construct the skewed generator.
         */
        skewed_generator(const value_type& min, const value_type& max, generator_type& generator) :
                shuffled(generate_likeliest_values(min, max)), generator(generator) { }


    private:
        /**
         * Order the possible values based on likelihood.
         */
        static std::vector<value_type> generate_likeliest_values(const value_type& min, const value_type& max)
        {
            std::vector<value_type> result;
            for (value_type i = min; i <= max; ++i) {
                result.emplace_back(i);
            }
            std::shuffle(result.begin(), result.end(), std::random_device());
            return result;
        }
    };


    /**
     * Return a skewed generator.
     */
    template<class Type, class Generator>
    skewed_generator<Type, Generator> make_skewed_generator(const Type& min, const Type& max, Generator& generator)
    {
        return skewed_generator<Type, Generator>(min, max, generator);
    }
};

#endif